import sys
from typing import Optional

import opengate_core as g4
from ..base import GateObject, process_cls
from ..exception import fatal


class FilterBase(GateObject):
    """
    A filter to be attached to an actor.
    """

    # hints for IDE
    policy: str

    user_info_defaults = {
        "policy": (
            "accept",
            {
                "doc": "How should the item be handled?",
                "allowed_values": ["accept", "reject"],
            },
        ),
    }

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def __initcpp__(self):
        """Nothing to do in the base class."""

    def initialize(self):
        self.InitializeUserInfo(self.user_info)

    def __setstate__(self, state):
        self.__dict__ = state
        self.__initcpp__()


class ParticleFilter(FilterBase, g4.GateParticleFilter):

    # hints for IDE
    particle: str

    user_info_defaults = {
        "particle": (
            "",
            {
                "doc": "Name of the particle to which this filter is applied.",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        FilterBase.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateParticleFilter.__init__(self)


class KineticEnergyFilter(FilterBase, g4.GateKineticEnergyFilter):

    # hints for IDE
    energy_min: float
    energy_max: float

    user_info_defaults = {
        "energy_min": (
            0,
            {
                "doc": "Lower kinetic energy bound.",
            },
        ),
        "energy_max": (
            sys.float_info.max,
            {
                "doc": "Upper kinetic energy bound.",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        FilterBase.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateKineticEnergyFilter.__init__(self)  # no argument in cpp side


class TrackCreatorProcessFilter(FilterBase, g4.GateTrackCreatorProcessFilter):

    # hints for IDE
    process_name: str

    user_info_defaults = {
        "process_name": (
            "none",
            {
                "doc": "Name of the track creator process to be identified.",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        FilterBase.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateTrackCreatorProcessFilter.__init__(self)  # no argument in cpp side


class ThresholdAttributeFilter(FilterBase, g4.GateThresholdAttributeFilter):

    # hints for IDE
    value_min: float
    value_max: float
    attribute: Optional[str]

    user_info_defaults = {
        "value_min": (
            0,
            {
                "doc": "Lower bound. ",
            },
        ),
        "value_max": (
            sys.float_info.max,
            {
                "doc": "Upper bound. ",
            },
        ),
        "attribute": (
            None,
            {
                "doc": "Attribute name to be considered. ",
            },
        ),
    }

    # FIXME required test dans initialize

    def __init__(self, *args, **kwargs):
        FilterBase.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateThresholdAttributeFilter.__init__(self)

    def initialize(self):
        if self.attribute is None:
            fatal(
                f"The user input parameter 'attribute' is not set but required in filter '{self.name}'."
            )
        super().initialize()


class UnscatteredPrimaryFilter(FilterBase, g4.GateUnscatteredPrimaryFilter):

    def __init__(self, *args, **kwargs):
        FilterBase.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateUnscatteredPrimaryFilter.__init__(self)

    def initialize(self):
        super().initialize()


filter_classes = {
    "ParticleFilter": ParticleFilter,
    "KineticEnergyFilter": KineticEnergyFilter,
    "TrackCreatorProcessFilter": TrackCreatorProcessFilter,
    "ThresholdAttributeFilter": ThresholdAttributeFilter,
    "UnscatteredPrimaryFilter": UnscatteredPrimaryFilter,
}


def get_filter_class(f):
    try:
        return filter_classes[f]
    except KeyError:
        fatal(
            f"Unknown filter '{f}'. Known filters are: {list(filter_classes.keys())}."
        )


process_cls(FilterBase)
process_cls(ParticleFilter)
process_cls(KineticEnergyFilter)
process_cls(TrackCreatorProcessFilter)
process_cls(ThresholdAttributeFilter)
process_cls(UnscatteredPrimaryFilter)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.tests.utility as utility
from opengate.utility import g4_units

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test004_simulation_stats_actor"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.running_verbose_level = 0
    sim.visu = False
    sim.number_of_threads = 5
    # special debug mode : force MT even with one single thread
    sim.force_multithread_mode = True
    sim.random_engine = "MixMaxRng"
    sim.random_seed = "auto"

    """
        Warning: we can only see the speed up of the MT mode for large nb of particles (>2e6)
    """

    # set the world size like in the Gate macro
    m = g4_units.m
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_AIR"

    # add a simple waterbox volume
    waterbox = sim.add_volume("Box", "Waterbox")
    cm = g4_units.cm
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
    waterbox.material = "G4_WATER"

    # physic list
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMV"
    um = g4_units.um
    global_cut = 700 * um
    sim.physics_manager.global_production_cuts.gamma = global_cut
    sim.physics_manager.global_production_cuts.electron = global_cut
    sim.physics_manager.global_production_cuts.positron = global_cut
    sim.physics_manager.global_production_cuts.proton = global_cut

    # default source for tests
    keV = g4_units.keV
    mm = g4_units.mm
    Bq = g4_units.Bq
    source = sim.add_source("GenericSource", "Default")
    source.particle = "gamma"
    source.energy.mono = 80 * keV
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.n = 200000 / sim.number_of_threads

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # start simulation
    # sim.g4_commands_after_init.append("/run/verbose 0")
    # sim.g4_commands_after_init.append("/run/eventModulo 5000 1")
    sim.run()

    # get results
    print(stats)
    print("track type", stats.counts.track_types)

    # gate_test4_simulation_stats_actor
    # Gate mac/main.mac
    stats_ref = utility.read_stat_file(paths.gate_output / "stat.txt")
    stats_ref.counts.runs = sim.number_of_threads
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.01)

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Context: See test079_acollin_ions_geant4_material.py

Here, the source is between two materials.
In the first simulation, only one of them has the correct energy per ion for acollin.
Thus, the test expect a dirac and a Rayleigh distributions.
In the second simulation, both materials have the correct energy per ion for acollin.
Thus. the test expect a Rayleigh distribution.
"""

from test079_acollin_helpers import *
import opengate.tests.utility as tu
import matplotlib.pyplot as plt


#########################################################################################
# Simulations configuration that may be relevant to change
#########################################################################################
# Mean energy of Ion Pair to use. 5.0 eV should produce the expected 0.5 deg FWHM in PET
# imaging
mean_energy = 5.0 * eV
# Key added to output to make sure that multi-threading the tests does not backfire
test_key = "p6"


#########################################################################################
# Main : We use this to launch the test
#########################################################################################
if __name__ == "__main__":
    paths = tu.get_default_test_paths(__file__, output_folder="test079")

    # Define core of the simulation, including physics
    sim = setup_simulation_engine(paths)

    # add a volume that englobe the two objets for the actor
    wb0 = sim.add_volume("Box", "both_obj")
    e = 0.01 * cm
    wb0.size = [e + 50 * cm, e + 100 * cm, e + 50 * cm]
    wb0.material = "G4_AIR"

    # add a bodybox
    wb1 = sim.add_volume("Box", "bodybox")
    wb1.mother = wb0
    wb1.size = [50 * cm, 50 * cm, 50 * cm]
    wb1.translation = [0 * cm, -25.0001 * cm, 0 * cm]
    wb1.material = "Body"

    # add a waterbox
    wb2 = sim.add_volume("Box", "waterbox")
    wb2.mother = wb0
    wb2.size = [50 * cm, 50 * cm, 50 * cm]
    wb2.translation = [0 * cm, 25.0001 * cm, 0 * cm]
    wb2.material = "Water"

    # set the source
    source = sim.add_source("GenericSource", "f18")
    source.particle = "ion 9 18"
    source.energy.mono = 0
    source.activity = 10000 * Bq
    source.direction.type = "iso"

    # add phase actor
    phsp = setup_actor(sim, "phsp", wb0.name)
    phsp.output_filename = paths.output / f"annihilation_photons_{test_key}.root"

    sim.physics_manager.mean_energy_per_ion_pair["Water"] = mean_energy

    # go
    sim.run(start_new_process=True)

    # redo test changing the MeanEnergyPerIonPair
    root_filename1 = phsp.output_filename
    phsp.output_filename = (
        paths.output / f"annihilation_photons_with_mepip_{test_key}.root"
    )

    sim.physics_manager.mean_energy_per_ion_pair["Body"] = mean_energy

    # go
    sim.run()

    # test: no mean energy, should be mostly collinear
    gamma_pairs = read_gamma_pairs(root_filename1, "phsp")
    acollinearity_angles = compute_acollinearity_angles(gamma_pairs)

    plt.hist(
        acollinearity_angles,
        bins=71,
        range=(0, 1.0),
        alpha=0.7,
        color="blue",
        label=f"Only one material with mean energy per Ion par of {mean_energy / eV:.1f} eV ",
    )
    # Significant, like 30%, have angle bigger than 0.8 deg... seems
    # to be due to curr_actor.steps_to_store = "first". Not sure how
    # to make the test more clean...
    ratio_colin = np.sum(np.array(acollinearity_angles) < 0.01) / np.sum(
        np.array(acollinearity_angles) < 0.8
    )
    print(f"Ratio of collinear annihilation pairs: {ratio_colin}")

    # test: with mean energy, acolinearity amplitude should have a Rayleigh distribution
    gamma_pairs = read_gamma_pairs(phsp.output_filename, "phsp")
    acollinearity_angles = compute_acollinearity_angles(gamma_pairs)

    acolin_scale = plot_acolin_case_mepip(mean_energy, acollinearity_angles)

    f = paths.output / f"acollinearity_angles_{test_key}.png"
    plt.savefig(f)
    print(f"Plot was saved in {f}")

    # final
    # No acolin
    is_ok_p1 = np.isclose(ratio_colin, 0.5, atol=0.1)
    # Basic acolin
    is_ok_p2 = np.isclose(acolin_scale * 2.355, 0.5, atol=0.2)

    tu.test_ok(is_ok_p1 and is_ok_p2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test053_phid_helpers2 import *
import opengate as gate


if __name__ == "__main__":
    paths = get_default_test_paths(__file__, "", output_folder="test053")

    # bi213 83 213
    # ac225 89 225
    # fr221 87 221
    # lu177 71 177
    z = 89
    a = 225
    nuclide, _ = get_nuclide_and_direct_progeny(z, a)
    print(nuclide)

    sim = gate.Simulation()
    sim_name = f"{nuclide.nuclide}_5_ref"
    create_sim_test053(sim, sim_name)

    # sources
    sim.number_of_threads = 8
    activity_in_Bq = 1000
    add_source_generic(sim, z, a, activity_in_Bq)

    # timing
    sec = g4_units.second
    min = g4_units.minute
    start_time = 0 * min
    end_time = start_time + 20 * min
    duration = end_time - start_time
    print(f"start time {start_time / sec}")
    print(f"end time {end_time / sec}")
    print(f"Duration {duration / sec}")
    print(f"Ions {activity_in_Bq * duration / sec:.0f}")
    sim.run_timing_intervals = [[0, end_time]]

    # go
    sim.run()

    # print stats
    stats = sim.get_actor("stats")
    print(stats)

    # compare with reference root file
    warning(f"check root files")
    root_model = sim.get_actor("phsp").get_output_path()
    root_ref = paths.output_ref / Path(root_model).name
    keys = ["KineticEnergy", "TrackCreatorModelIndex"]
    tols = [0.001, 0.02]
    img = paths.output / str(root_model).replace(".root", ".png")
    is_ok = compare_root3(
        root_ref, root_model, "phsp", "phsp", keys, keys, tols, None, None, img
    )

    test_ok(is_ok)



from anytree import RenderTree
import numpy as np
from scipy.spatial.transform import Rotation

import opengate_core as g4
from ..definitions import __world_name__
from ..exception import fatal

"""
http://geant4-userdoc.web.cern.ch/geant4-userdoc/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html#constructed-solid-geometry-csg-solids
"""


def box_add_size(box, thickness):
    box.size = [x + thickness for x in box.size]


def cons_add_size(cons, thickness):
    cons.rmax1 += thickness / 2
    cons.rmax2 += thickness / 2
    cons.dz += thickness


def translate_point_to_volume(simulation, volume, top, x):
    """

    Consider the point x in the current volume and return the coordinate of x in the top volume
    (that must be an ancestor).
    Translation only, do not consider rotation.
    """
    while volume.name != top:
        x += volume.translation
        volume = simulation.volume_manager.volumes[volume.mother]
    return x


def render_tree(tree, geometry, world_name):
    """
    Print a tree of volume
    """
    s = ""
    for pre, fill, node in RenderTree(tree[world_name]):
        v = geometry[node.name]
        s += f"{pre}{node.name} {v.type_name} {v.material}\n"

    # remove last break line
    return s[:-1]


"""
A rotation matrix (3x3) can be represented by:
- G4RotationMatrix in module opengate_core
- np.array in module numpy
- Rotation in module scipy.spatial.transform

With scipy and np:
- rot_np = rot_scipy.as_matrix()
- rot_scipy = Rotation.from_matrix(rot_np)

With G4RotationMatrix
- rot_g4 = rot_np_as_g4(rot_np)
- rot_np = rot_g4_as_np(rot_g4)

Also for G4ThreeVector
- v_np = vec_g4_as_np(v_g4)
- v_g4 = vec_np_as_g4(v_np)
"""


def is_rotation_matrix(R):
    """
    https://stackoverflow.com/questions/53808503/how-to-test-if-a-matrix-is-a-rotation-matrix
    """
    # square matrix test
    if R.ndim != 2 or R.shape[0] != R.shape[1]:
        return False
    should_be_identity = np.allclose(R.dot(R.T), np.identity(R.shape[0], np.float64))
    should_be_one = np.allclose(np.linalg.det(R), 1)
    return should_be_identity and should_be_one


def vec_np_as_g4(v):
    return g4.G4ThreeVector(v[0], v[1], v[2])


def vec_g4_as_np(v):
    vnp = np.zeros(3)
    vnp[0] = v.x
    vnp[1] = v.y
    vnp[2] = v.z
    return vnp


def rot_np_as_g4(rot):
    if not is_rotation_matrix(rot):
        fatal(f"This matrix is not a rotation matrix (not orthogonal): \n{rot}")
    try:
        r = g4.HepRep3x3(
            rot[0, 0],
            rot[0, 1],
            rot[0, 2],
            rot[1, 0],
            rot[1, 1],
            rot[1, 2],
            rot[2, 0],
            rot[2, 1],
            rot[2, 2],
        )
    except Exception as e:
        s = f"Cannot convert the rotation {rot} to a 3x3 matrix. Exception is: "
        s += str(e)
        fatal(s)
    a = g4.G4RotationMatrix()
    a.set(r)
    return a


def rot_g4_as_np(rot):
    r = np.zeros(shape=(3, 3))
    r[0, 0] = rot.xx()
    r[0, 1] = rot.xy()
    r[0, 2] = rot.xz()
    r[1, 0] = rot.yx()
    r[1, 1] = rot.yy()
    r[1, 2] = rot.yz()
    r[2, 0] = rot.zx()
    r[2, 1] = rot.zy()
    r[2, 2] = rot.zz()
    if not is_rotation_matrix(r):
        fatal(f"The G4 matrix is not a rotation matrix (not orthogonal): \n{rot}")
    return r


def ensure_is_g4_translation(translation):
    if isinstance(translation, g4.G4ThreeVector):
        return translation
    else:
        return vec_np_as_g4(translation)


def ensure_is_g4_rotation(rotation):
    if isinstance(rotation, g4.G4RotationMatrix):
        return rotation
    else:
        return rot_np_as_g4(rotation)


def ensure_is_g4_transform(
    translation=(0, 0, 0), rotation=Rotation.identity().as_matrix()
):
    return g4.G4Transform3D(
        ensure_is_g4_rotation(rotation), ensure_is_g4_translation(translation)
    )


def get_translation_from_rotation_with_center(rot, center):
    center = np.array(center)
    t = rot.apply(-center) + center
    # note: apply is the same than rot.as_matrix().dot()
    return t


def get_transform_orbiting(initial_position, axis, angle_deg):
    angle_deg = list([angle_deg])
    translations = []
    rotations = []
    for ang in angle_deg:
        rot = Rotation.from_euler(axis, ang, degrees=True)
        t = rot.apply(np.array(initial_position))
        translations.append(t)
        rotations.append(rot.as_matrix())
    if len(translations) > 1:
        return translations, rotations
    else:
        return translations[0], rotations[0]


def get_transform_world_to_local(volume, repetition_index=None):
    """Calculate the rotation and translation needed
    to transform from the world reference frame
    into the local reference frame of this volume.

    If repetition_index is None:
    Returns two lists, the first with translation vectors,
    the second with rotation matrices. Each list entry corresponds
    to one physical volume of the Gate volume, i.e. one repetition.
    For non-repeated volumes, the lists will contain one item only.

    If repetition_index is a valid integer,
    return the translation and rotation for that repeated physical volume only.
    """

    volume._request_volume_tree_update()

    cumulative_translation = []
    cumulative_rotation = []
    # Note: access translation and rotation via user_info dictionary
    for i in range(volume.number_of_repetitions):
        ctr = volume.translation_list[i]
        crot = volume.rotation_list[i]
        for vol in volume.ancestor_volumes[::-1]:
            # in parallel world, the rotation_list does not exist
            if hasattr(vol, "rotation_list"):
                crot = np.matmul(vol.rotation_list[0], crot)
                ctr = vol.rotation_list[0].dot(ctr) + vol.translation_list[0]
        cumulative_translation.append(ctr)
        cumulative_rotation.append(crot)

    if repetition_index is None:
        return cumulative_translation, cumulative_rotation
    else:
        return (
            cumulative_translation[repetition_index],
            cumulative_rotation[repetition_index],
        )


def get_transform_world_to_local_old(vol_name):
    # cumulated translation and rotation
    ctr = [0, 0, 0]
    crot = Rotation.identity().as_matrix()
    first = True
    # FIXME for parallel world
    while vol_name != __world_name__:
        pv = g4.G4PhysicalVolumeStore.GetInstance().GetVolume(vol_name, False)
        tr = vec_g4_as_np(pv.GetObjectTranslation())
        rot = rot_g4_as_np(pv.GetObjectRotation())
        if first:
            ctr = tr
            crot = rot
            first = False
        else:
            crot = np.matmul(rot, crot)
            ctr = rot.dot(ctr) + tr

        if pv.GetMotherLogical() is None:
            vol_name = __world_name__
        else:
            vol_name = pv.GetMotherLogical().GetName()

    return ctr, crot


def get_circular_repetition(
    number_of_repetitions,
    first_translation,
    angular_step_deg="auto_full_circle",
    start_angle_deg=0.0,
    additional_rotation=Rotation.identity().as_matrix(),
    axis=(0, 0, 1),
):
    """Generate translations and rotations to repeat volumes in a circle.

    This helper function generates translations and rotations for a volume to be repeated in a circle,
    e.g. in a PET ring. The returned lists with translations and rotations can be used as input to the translation
    and rotation parameter of any repeatable volume in Gate.

    Args:
        number_of_repetitions (int) : How many times should the volume be repeated?
        first_translation (3-vector) : Where should the first copy of the volume be placed (wrt. to the mother volume)?
        angular_step_deg (float, optional) : The angular step in degrees between subsequent repetitions.
            Accepts a number or two special arguments, 'auto_full_circle' and 'auto_half_circle',
            to determine the angular step automatically. Default: 'auto_full_circle'
        start_angle_deg (int, optional) : The angle at which the repetition starts.
            The first volume copy is placed at `first_translation` and then rotated by `start_angle_deg`.
            Default: 0.
        additional_rotation (3x3 rotation matrix, optional) : Additional rotation to be applied to all copies,
            e.g. if the volume is tilted. Default: 3x3 identity.
        axis (3-vector, optional) : The axis (in the mother's frame of reference) around which
            the circular repetition is performed. Default: [0, 0, 1], i.e. z-axis, circle in the x-y-plane.

    Returns:
        list : A list of translation vectors, one for each repetition.
        list : A list of rotation matrices, one for each repetition.
    """
    if not is_rotation_matrix(additional_rotation):
        fatal(f"Invalid rotation matrix 'additional_rotation': {additional_rotation}.")

    if angular_step_deg == "auto_full_circle":
        angular_step_deg = 360.0 / number_of_repetitions
    elif angular_step_deg == "auto_half_circle":
        angular_step_deg = 180.0 / number_of_repetitions
    elif not isinstance(angular_step_deg, (int, float)):
        fatal(
            f"The input variable 'angular_step_deg' should be a number (int, float) "
            f"or one of the following terms 'auto_full_circle', 'auto_half_circle'. "
            f"Received: {angular_step_deg} which is of type {type(angular_step_deg).__name__}. "
        )
    angular_step_rad = np.deg2rad(angular_step_deg)
    start_angle_rad = np.deg2rad(start_angle_deg)
    translations = []
    rotations = []
    for angle in np.arange(
        start_angle_rad,
        start_angle_rad + number_of_repetitions * angular_step_rad,
        angular_step_rad,
    ):
        rot = Rotation.from_rotvec(angle * np.array(axis))
        rotations.append(rot.as_matrix().dot(additional_rotation))
        translations.append(rot.apply(first_translation))

    return translations, rotations


def get_grid_repetition(size, spacing, start=None, return_lut=False):
    """Generate a list of 3-vectors to be used as 'translation' parameter of a repeated volume.

    Args:
        size (list, np.ndarray) : 3-item list or numpy array specifying the number of repetitions
            along the axes x, y, z.
        spacing (list, np.ndarray) : 3-item list or numpy array specifying the spacing along the axes x, y, z
            between the translation vectors.
        start (optional): Optional 3-item list or numpy array specifying the first translation vector on the grid.
            If not provided, the grid is centered around (0,0,0).
        return_lut (bool, optional) : If true, the functions also returns a dictionary mapping copy index
            to the respective translation vector for later reference.

    Returns:
        list : A list of translations vectors.
        dict : (Optional) A dictionary mapping copy index to the respective translation vector. Only if `return_lut` is `True`.
    """
    if len(size) != 3:
        fatal(
            f"Input `size` must be a 3-item list or numpy array. Found length {len(size)}."
        )
    if len(spacing) != 3:
        fatal(
            f"Input `spacing` must be a 3-item list or numpy array. Found length {len(spacing)}."
        )

    size = np.asarray(size)
    spacing = np.asarray(spacing)

    if start is None:
        start = -(size - 1) * spacing / 2.0
    translations = [start + spacing * np.array(pos) for pos in np.ndindex(*size)]

    if return_lut is True:
        lut = dict([(i, tr) for i, tr in enumerate(translations)])
        return translations, lut
    else:
        return translations


def volume_orbiting_transform(axis, start, end, n, initial_t, initial_rot):
    angle = start
    step_angle = (end - start) / n
    translations = []
    rotations = []
    for r in range(n):
        irot = Rotation.from_matrix(initial_rot)
        t, rot = get_transform_orbiting(initial_t, axis, angle)
        rot = Rotation.from_matrix(rot)
        rot = rot * irot
        translations.append(t)
        r = rot.as_matrix()
        rotations.append(r)
        angle += step_angle
    return translations, rotations



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from scipy.spatial.transform import Rotation
import opengate.contrib.phantoms.nemaiec as gate_iec

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test015")

    # create the simulation
    sim = gate.Simulation()

    # units
    MeV = gate.g4_units.MeV
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    cm3 = gate.g4_units.cm3
    Bq = gate.g4_units.Bq
    BqmL = Bq / cm3

    # main options
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = True
    sim.random_seed = 123654987
    sim.output_dir = paths.output

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.set_production_cut("world", "all", 10 * mm)

    # world size
    sim.world.size = [0.5 * m, 0.5 * m, 0.5 * m]

    # add an iec phantom
    # rotation 180 around X to be like in the iec 61217 coordinate system
    iec_phantom = gate_iec.add_iec_phantom(sim)
    iec_phantom.translation = [1 * cm, 2 * cm, 3 * cm]
    iec_phantom.rotation = Rotation.from_euler("x", 180, degrees=True).as_matrix()

    # add sources for all spheres
    a = 100 * BqmL
    activity_Bq_mL = [10 * a, 2 * a, 3 * a, 4 * a, 5 * a, 6 * a]
    sources = gate_iec.add_spheres_sources(
        sim, iec_phantom.name, "sources", "all", activity_Bq_mL, verbose=True
    )
    for source in sources:
        source.particle = "alpha"
        source.energy.type = "mono"
        source.energy.mono = 100 * MeV

    # add stat actor
    stats_actor = sim.add_actor("SimulationStatisticsActor", "stats")
    stats_actor.track_types_flag = True
    stats_actor.output_filename = "test015_iec_2_stats.txt"

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.edep.output_filename = "test015_iec_2.mhd"
    dose.attached_to = iec_phantom.name
    dose.size = [100, 100, 100]
    dose.spacing = [2 * mm, 2 * mm, 2 * mm]

    # start
    sim.run()

    # compare stats
    stats_ref = utility.read_stat_file(paths.output_ref / "test015_iec_2_stats.txt")
    is_ok = utility.assert_stats(stats_actor, stats_ref, tolerance=0.03)

    # compare images
    im_ok = utility.assert_images(
        paths.output_ref / "test015_iec_2.mhd",
        dose.edep.get_output_path(),
        stats_actor,
        tolerance=40,
        ignore_value_data2=0,
        axis="x",
        sum_tolerance=2,
        apply_ignore_mask_to_sum_check=False,  # force legacy behavior
    )

    is_ok = is_ok and im_ok
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test022_half_life_helpers as test022
import opengate as gate
from opengate.tests import utility
import math
import sys
import matplotlib.pyplot as plt

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test022")

    # create the simulation
    sim = gate.Simulation()

    # multithread ?
    argv = sys.argv
    n = 1
    if len(argv) > 1:
        n = int(argv[1])

    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.number_of_threads = n
    sim.random_seed = 92344321
    sim.output_dir = paths.output
    print(sim)

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    nm = gate.g4_units.nm
    keV = gate.g4_units.keV
    Bq = gate.g4_units.Bq
    sec = gate.g4_units.s

    # set the world size like in the Gate macro
    sim.world.size = [10 * m, 10 * m, 10 * m]

    # waterbox
    waterbox1 = sim.add_volume("Box", "waterbox1")
    waterbox1.size = [50 * cm, 50 * cm, 50 * cm]
    waterbox1.translation = [50 * cm, 0 * cm, 0 * cm]
    waterbox1.material = "G4_WATER"

    # plane between the two waterbox to stop gamma
    gcm3 = gate.g4_units.g_cm3
    sim.volume_manager.material_database.add_material_nb_atoms(
        "Tung", ["W"], [1], 1000 * gcm3
    )
    tung_plane = sim.add_volume("Box", "tung_plane")
    tung_plane.size = [1 * cm, 300 * cm, 300 * cm]
    tung_plane.translation = [0 * cm, 0 * cm, 0 * cm]
    tung_plane.material = "Tung"

    # waterbox
    waterbox2 = sim.add_volume("Box", "waterbox2")
    waterbox2.size = [50 * cm, 50 * cm, 50 * cm]
    waterbox2.translation = [-50 * cm, 0 * cm, 0 * cm]
    waterbox2.material = "G4_WATER"

    # physics
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMZ"
    sim.physics_manager.enable_decay = True
    sim.physics_manager.global_production_cuts.all = (
        1 * mm
    )  # all means: proton, electron, positron, gamma

    # activity
    activity_Bq = 8000 * Bq
    half_life = 5 * sec
    lifetime = half_life / math.log(2.0)

    # "hole" in the timeline to check if no particle are emitted at this moment
    sim.run_timing_intervals = [[0 * sec, 12 * sec], [13 * sec, 20 * sec]]

    # source #1
    source1 = sim.add_source("GenericSource", "source1")
    source1.attached_to = "waterbox1"
    source1.particle = "ion 49 111"  # In111 171 keV and 245 keV
    source1.position.type = "sphere"
    source1.position.radius = 1 * mm
    source1.direction.type = "iso"
    source1.activity = activity_Bq / sim.number_of_threads
    source1.half_life = half_life
    # this is needed, but automatically done in GenericSource.py
    source1.user_particle_life_time = 0
    print()
    print(f"Source1 ac = {source1.activity / Bq} Bq")
    print(f"Source1 HL = {half_life / sec} sec")

    # source #2
    source2 = sim.add_source("GenericSource", "source2")
    source2.attached_to = "waterbox2"
    source2.particle = "ion 49 111"  # In111 171 keV and 245 keV
    source2.position.type = "sphere"
    source2.position.radius = 1 * mm
    source2.position.translation = [0, 0, -3 * cm]
    source2.direction.type = "iso"
    source2.user_particle_life_time = lifetime
    source2.n = activity_Bq / Bq / sim.number_of_threads * lifetime / sec
    print()
    print("Source2 n = ", source2.n)
    print(f"Source2 HL = {half_life / sec} sec")
    print(f"Source2 LT = {lifetime / sec} sec")
    print()

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # hit actor w1
    ta1 = sim.add_actor("PhaseSpaceActor", "PhaseSpace1")
    ta1.attached_to = "waterbox1"
    ta1.attributes = ["KineticEnergy", "GlobalTime", "PreGlobalTime"]
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = "gamma"
    f.policy = "accept"
    ta1.filters.append(f)
    ta1.output_filename = "test022_half_life_ion1.root"
    ta1.steps_to_store = "first"

    # hit actor w2
    ta2 = sim.add_actor("PhaseSpaceActor", "PhaseSpace2")
    ta2.attached_to = "waterbox2"
    ta2.attributes = ["KineticEnergy", "GlobalTime", "PreGlobalTime"]
    ta2.filters.append(f)
    ta2.output_filename = "test022_half_life_ion2.root"
    ta2.steps_to_store = "first"

    # start simulation
    sim.run()

    # get result
    stats = sim.get_actor("Stats")
    print(stats)

    # tests
    fig, ax = plt.subplots(ncols=2, nrows=1, figsize=(15, 5))
    b1, n1 = test022.test_half_life_fit(sim, ta1.get_output_path(), half_life, ax[0])
    b2, n2 = test022.test_half_life_fit(sim, ta2.get_output_path(), half_life, ax[1])
    fn = paths.output / "test022_half_life_ion_fit.png"
    print("Figure in ", fn)
    plt.savefig(fn)
    is_ok = b1 and b2

    # compare number of gammas
    diff = math.fabs(n1 - n2) / n2
    tol = 0.05
    b = diff < tol
    print()
    utility.print_test(
        b, f"Number of emitted gammas {n1} vs {n2} : {diff*100:.2f} % (tol is {tol})"
    )
    is_ok = is_ok and b

    utility.test_ok(is_ok)



from .exception import fatal, warning


class UserElement:
    """
    Common class for all types of elements (volume, source or actor)
    Manager a dict (Box) for user parameters: user_info
    Check that all the required keys are provided
    """

    def __init__(self, user_info):
        # set the user info (a kind of dict)
        # check everything is there (except for solid building)
        check_user_info(user_info)  # removed to avoid circular import
        # self.check_user_info()
        self.user_info = user_info
        # check type_name
        if self.user_info.type_name != self.type_name:
            fatal(
                f"Error, the type_name inside the user_info is different "
                f"from the type_name of the class: {self.user_info} in the "
                f"class {self.__name__} {self.type_name}"
            )
        # by default the name is a unique id (uuid)
        if not self.user_info.name:
            fatal(
                f"Error a {self.user_info.volume_type} must have "
                f"a valid name, while it is {self.user_info.name}"
            )
        # debug
        self.verbose_getstate = False
        self.verbose_close = False

    @staticmethod
    def set_default_user_info(user_info):
        # Should be overwritten by subclass
        pass

    def __str__(self):
        s = f"Element: {self.user_info}"
        return s

    def set_simulation(self, simulation):
        self.simulation = simulation
        if simulation is not None:
            self.verbose_getstate = self.simulation.verbose_getstate
            self.verbose_close = self.simulation.verbose_close


def check_user_info(user_info):
    # get a fake ui to compare
    from .userinfo import UserInfo

    ref_ui = UserInfo(user_info.element_type, user_info.type_name)
    for val in ref_ui.__dict__:
        if val not in user_info.__dict__:
            fatal(f'Cannot find "{val}" in {user_info}')
    for val in user_info.__dict__:
        if val not in ref_ui.__dict__.keys():
            warning(f'Unused param "{val}" in {user_info}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.userhooks import user_hook_active_regions, user_hook_em_switches
from opengate.tests import utility


def combined_user_hook_after_run(simulation_engine):
    user_hook_em_switches(simulation_engine)
    user_hook_active_regions(simulation_engine)


def check_hook_output(sim):
    # Check if G4 set flags correctly
    print(sim.user_hook_log[0])
    print(sim.user_hook_log[1])
    em_parameters_from_hook = sim.user_hook_log[0]
    for k, v in sim.physics_manager.em_parameters.items():
        assert v == em_parameters_from_hook[k]

    em_regions_from_hook = sim.user_hook_log[1]
    for r in sim.physics_manager.regions.values():
        assert r.em_switches.deex == em_regions_from_hook[r.name][0]
        assert r.em_switches.auger == em_regions_from_hook[r.name][1]
    assert (
        sim.physics_manager.em_switches_world.deex == em_regions_from_hook["world"][0]
    )
    assert (
        sim.physics_manager.em_switches_world.auger == em_regions_from_hook["world"][1]
    )


if __name__ == "__main__":
    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = True
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_engine = "MersenneTwister"
    sim.random_seed = 1234

    # shortcuts for units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    eV = gate.g4_units.eV
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq

    # set the world size
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]

    # add a simple waterbox volume
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
    waterbox.material = "G4_WATER"

    # add a daughter (in wb)
    b1 = sim.add_volume("Box", "b1")
    b1.mother = "waterbox"
    b1.size = [4 * cm, 4 * cm, 4 * cm]
    b1.translation = [5 * cm, 5 * cm, 0 * cm]
    b1.material = "G4_Pd"

    # add another box (in world)
    b2 = sim.add_volume("Box", "b2")
    b2.size = [4 * cm, 4 * cm, 4 * cm]
    b2.translation = [0 * cm, 0 * cm, 0 * cm]
    b2.material = "G4_LUNG_ICRP"

    # print info about physics
    print("Physics manager:\n", sim.physics_manager)
    print("Available phys lists:")
    print(sim.physics_manager.dump_available_physics_lists())

    source = sim.add_source("GenericSource", "gamma")
    source.particle = "gamma"
    source.energy.mono = 10 * MeV
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 10 * Bq  # do not need high stats, testv just checks flags

    # change physics
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMZ"

    # em parameters
    sim.physics_manager.em_parameters.fluo = True
    sim.physics_manager.em_parameters.auger = True
    sim.physics_manager.em_parameters.auger_cascade = True
    sim.physics_manager.em_parameters.pixe = True
    sim.physics_manager.em_parameters.deexcitation_ignore_cut = True

    # activate deex and auger in regions
    sim.physics_manager.em_switches_world.deex = True
    sim.physics_manager.em_switches_world.auger = True
    sim.physics_manager.em_switches_world.pixe = True

    region_b1 = sim.physics_manager.add_region("region_b1")
    region_b1.em_switches.deex = True
    region_b1.em_switches.auger = False
    region_b1.associate_volume(b1)

    region_b2 = sim.physics_manager.add_region("region_b2")
    region_b2.em_switches.deex = False
    region_b2.em_switches.auger = True
    region_b2.associate_volume(b2)

    sim.user_hook_after_run = combined_user_hook_after_run

    sim.run()
    check_hook_output(sim)

    utility.test_ok(True)



import copy
import inspect
from box import Box
from typing import Optional
import sys

import opengate_core as g4
from ..base import GateObject, process_cls
from ..utility import insert_suffix_before_extension, ensure_filename_is_str
from ..exception import warning, fatal, GateImplementationError
from .dataitems import (
    SingleItkImage,
    SingleMeanItkImage,
    SingleItkImageWithVariance,
    QuotientItkImage,
    QuotientMeanItkImage,
    merge_data,
)


def get_formatted_docstring_rst(cls, attr_name, begin_of_line="  - "):
    """Format a property's docstring into a reST bullet point."""
    attr = getattr(cls, attr_name, None)
    if not attr:
        raise GateImplementationError(f"Attribute {attr_name} not found in {cls}.")

    docstring = inspect.getdoc(attr)
    prop_or_func = attr_name
    if inspect.isfunction(attr):
        prop_or_func += "()"

    if not docstring:
        return f"{begin_of_line} **{prop_or_func}**: No description available.\n\n"

    description = " ".join(line.strip() for line in docstring.splitlines())

    return f"{begin_of_line} **{prop_or_func}**: {description}\n\n"


class BaseUserInterfaceToActorOutput:
    # these attributes are known to the class
    # and should be treated differently by __getattr__() and __setattr__(),
    # namely they should be retrieved directly from __dict__
    # or written directly into __dict__ to avoid infinite recursion
    # IMPORTANT: a copy of this list needs to be defined also in __getattr__ and __setattr__
    _known_attributes = (
        "__setstate__",
        "__getstate__",
        "user_output_name",
        "interface_name",
        "belongs_to_actor",
        "_kwargs_for_interface_calls",
    )

    @classmethod
    def __get_docstring_attributes__(cls):
        docstring = ""
        docstring += get_formatted_docstring_rst(cls, "active")
        docstring += get_formatted_docstring_rst(cls, "output_filename")
        docstring += get_formatted_docstring_rst(cls, "write_to_disk")
        docstring += get_formatted_docstring_rst(cls, "keep_data_per_run")
        return docstring

    @classmethod
    def __get_docstring_methods__(cls):
        docstring = ""
        docstring += get_formatted_docstring_rst(cls, "get_output_path")
        docstring += get_formatted_docstring_rst(cls, "get_data")
        return docstring

    def __init__(
        self,
        belongs_to_actor,
        user_output_name,
        interface_name,
        kwargs_for_interface_calls=None,
        **kwargs,
    ):
        # Important: the attributes set here in the __init__ method need to be
        # listed in the _known_attributes class attribute
        # because the __setattr__ method needs that
        self.user_output_name = user_output_name
        self.interface_name = interface_name
        self.belongs_to_actor = belongs_to_actor
        if kwargs_for_interface_calls is None:
            self._kwargs_for_interface_calls = {}
        else:
            self._kwargs_for_interface_calls = kwargs_for_interface_calls

    def __getstate__(self):
        """
        For earlier python version (<3.11), __getstate__ may not be defined.
        We provide a simple workaround here to return a copy of the internal dict.
        """
        return_dict = self.__dict__.copy()
        # set 'belongs_to_actor' to None to avoid pickling circles; will be reset during unpickling
        return_dict["belongs_to_actor"] = None
        return return_dict

    @property
    def _user_output(self):
        return self.belongs_to_actor.user_output[self.user_output_name]

    @property
    def active(self):
        """Should the actor consider and score this output?"""
        try:
            return self._user_output.get_active(**self._kwargs_for_interface_calls)
        except NotImplementedError:
            raise AttributeError

    @active.setter
    def active(self, value):
        self._user_output.set_active(value, **self._kwargs_for_interface_calls)

    def get_output_path(self, which="merged", **kwargs):
        """Get the path (absolute) where GATE stores this output.
        Use the argument 'which' to specify whether you refer to the cumulative output
        of the entire simulation (which='merged'), or to a specific run,
        e.g. which=2 for run index 2 (run indices start at 0).
        """
        kwargs.update(self._kwargs_for_interface_calls)
        return self._user_output.get_output_path(which=which, **kwargs)

    def get_run_indices(self, **kwargs):
        kwargs.update(self._kwargs_for_interface_calls)
        return self._user_output.get_run_indices(**kwargs)

    def get_data(self, which="merged", **kwargs):
        """Get the data stored in this output, e.g. an ITK image.
        Use the argument 'which' to specify whether you refer to the cumulative output
        of the entire simulation (which='merged'), or to a specific run,
        e.g. which=2 for run index 2 (run indices start at 0).
        """
        kwargs.update(self._kwargs_for_interface_calls)
        return self._user_output.get_data(**kwargs)

    @property
    def write_to_disk(self):
        """Should this output be stored on disk?"""
        try:
            return self._user_output.get_write_to_disk(
                **self._kwargs_for_interface_calls
            )
        except NotImplementedError:
            raise AttributeError

    @write_to_disk.setter
    def write_to_disk(self, value):
        self._user_output.set_write_to_disk(value, **self._kwargs_for_interface_calls)

    @property
    def output_filename(self):
        """Output filename used for this output.
        An automatic suffix is appended for per-run data.
        You can also specify a relative path, i.e. relative to the simulation's output directory,
        e.g. output_file = Path('dose_output') / 'patient_dose.mhd'.
        """
        try:
            return self._user_output.get_output_filename(
                **self._kwargs_for_interface_calls
            )
        except NotImplementedError:
            raise AttributeError

    @output_filename.setter
    def output_filename(self, value):
        self._user_output.set_output_filename(value, **self._kwargs_for_interface_calls)

    @property
    def keep_data_per_run(self):
        """Should data be kept in memory for individual runs? If False, only the cumulative data is kept.
        Note: Not every kind of user output supports this, e.g. ROOT output cannot per stored on a per-run basis.
        """
        try:
            return self._user_output.get_keep_data_per_run(
                **self._kwargs_for_interface_calls
            )
        except NotImplementedError:
            raise AttributeError

    @keep_data_per_run.setter
    def keep_data_per_run(self, value):
        self._user_output.keep_data_per_run = value

    @property
    def suffix(self):
        """Specify the automatic suffix to be used for this output in case the output_filename is set
        via the actor for all output handled by the actor.
        The default item suffix is equal to the output name and there should be no need to change it.
        """
        try:
            return self._user_output.get_item_suffix(**self._kwargs_for_interface_calls)
        except NotImplementedError:
            raise AttributeError

    @suffix.setter
    def suffix(self, value):
        self._user_output.set_item_suffix(value, **self._kwargs_for_interface_calls)

    def __getattr__(self, item):
        # Recall: this method is called when python cannot otherwise
        # find the attribute in the instance. In this case, we try to find it
        # in the associated user_output to make the interface transparent

        # try to get known attributes directly from __dict__
        # to avoid infinite recursion
        if item in (
            "__setstate__",
            "__getstate__",
            "user_output_name",
            "interface_name",
            "belongs_to_actor",
            "_kwargs_for_interface_calls",
        ):
            try:
                return self.__dict__[item]
            except KeyError:
                raise AttributeError(f"Could not find known attribute {item}")
        # for the others, use the getattr() builtin
        _user_output = self.belongs_to_actor.user_output[self.user_output_name]
        try:
            return getattr(_user_output, item)
        except AttributeError:
            raise AttributeError(
                f"Tried to find {item} in user output {_user_output.name} "
                "and via the interface to it, but it is not there. "
            )

    def __setattr__(self, item, value):
        # if item in type(self).__dict__["_known_attributes"]:
        if item in (
            "user_output_name",
            "belongs_to_actor",
            "_kwargs_for_interface_calls",
        ):
            self.__dict__[item] = value
        else:
            try:
                super().__setattr__(item, value)
            except NotImplementedError:
                if item in self._user_output.user_info:
                    setattr(self._user_output, item, value)
                else:
                    fatal(
                        f"Unable to set value {value} for item {item}. "
                        "Make sure the actor and/or actor output support this parameter. "
                    )


class UserInterfaceToActorOutputUsingDataItemContainer(BaseUserInterfaceToActorOutput):

    def __init__(self, *args, item=0, **kwargs):
        super().__init__(*args, kwargs_for_interface_calls={"item": item}, **kwargs)


class UserInterfaceToActorOutputImage(UserInterfaceToActorOutputUsingDataItemContainer):

    @classmethod
    def __get_docstring_attributes__(cls):
        docstring = super().__get_docstring_attributes__()
        docstring += get_formatted_docstring_rst(cls, "image")
        return docstring

    @property
    def image(self):
        """Shortcut to the ITK image containing the cumulative result of this actor,
        e.g. the dose scored over the entire simulation. If you need to get the image
        corresponding to a certain run, use get_data(which=RUN_INDEX).
        For example: get_data(which=3) to get the image from run 3 (run indices start at 0).
        """
        return self._user_output.get_data(**self._kwargs_for_interface_calls)


def _setter_hook_belongs_to(self, belongs_to):
    if belongs_to is None:
        fatal("The belongs_to attribute of an ActorOutput cannot be None.")
    try:
        belongs_to_name = belongs_to.name
    except AttributeError:
        belongs_to_name = belongs_to
    return belongs_to_name


class ActorOutputBase(GateObject):
    # hints for IDE
    belongs_to: str
    keep_data_in_memory: bool

    _default_interface_class = BaseUserInterfaceToActorOutput
    default_suffix = None

    user_info_defaults = {
        "belongs_to": (
            None,
            {
                "doc": "Name of the actor to which this output belongs.",
                "setter_hook": _setter_hook_belongs_to,
                "required": True,
            },
        ),
        "keep_data_in_memory": (
            True,
            {
                "doc": "Should the data be kept in memory after the end of the simulation? "
                "Otherwise, it is only stored on disk and needs to be re-loaded manually. "
                "Careful: Large data structures like a phase space need a lot of memory.",
            },
        ),
    }

    @classmethod
    def get_default_interface_class(cls):
        if cls._default_interface_class is None:
            raise GateImplementationError(
                f"This class has no _default_interface_class class attribute defined. "
            )
        return cls._default_interface_class

    @classmethod
    def get_user_info_default_values_interface(cls, **kwargs):
        # FIXME: not sure yet how to handle keep_data_in_memory
        #        because it is not a per-interface but per actor output parameter
        # defaults = {"keep_data_in_memory": cls.inherited_user_info_defaults["keep_data_in_memory"][0]}
        defaults = {}
        return defaults

    @classmethod
    def set_user_info_default_values_interface(cls, **kwargs):
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.data_per_run = {}  # holds the data per run in memory
        self.merged_data = None  # holds the data merged from multiple runs in memory

    def __len__(self):
        return len(self.data_per_run)

    def get_run_indices(self, **kwargs):
        return [k for k, v in self.data_per_run.items() if v is not None]

    def set_write_to_disk(self, value, **kwargs):
        raise NotImplementedError

    def get_write_to_disk(self, **kwargs):
        raise NotImplementedError

    def set_output_filename(self, value, **kwargs):
        raise NotImplementedError

    def get_output_filename(self, **kwargs):
        raise NotImplementedError

    def get_active(self, **kwargs):
        # actor output is always active in its base implementation;
        # derived classes can implement this differently, as for example the ActorOutputUsingDataItemContainer class
        return True

    def set_active(self, value, **kwargs):
        if value is False:
            self.warn_user(
                f"You try to deactivate user output {self.name} "
                f"belonging to actor {self.belongs_to}, but this output cannot be deactivated. "
            )

    def get_item_suffix(self, **kwargs):
        return None

    def set_item_suffix(self, value, **kwargs):
        raise NotImplementedError

    @property
    def belongs_to_actor(self):
        if self.simulation is None:
            fatal(
                "Cannot determine the actor to which this output belongs. "
                "Probably, the actor has not yet been added to a simulation. "
            )
        return self.simulation.actor_manager.get_actor(self.belongs_to)

    def initialize(self):
        pass

    def _generate_auto_output_filename(self, **kwargs):
        return f"{self.name}_from_{self.belongs_to_actor.type_name.lower()}_{self.belongs_to_actor.name}.{self.default_suffix}"

    def _compose_output_path(self, which, output_filename):
        full_data_path = self.simulation.get_output_path(output_filename)

        if which == "merged":
            return full_data_path
        else:
            try:
                run_index = int(which)
            except ValueError:
                fatal(
                    f"Invalid argument 'which' in get_output_path() method "
                    f"of {type(self).__name__} called {self.name}"
                    f"Valid arguments are a run index (int) or the term 'merged'. "
                )
                run_index = None  # remove warning from IDE
            return insert_suffix_before_extension(full_data_path, f"run{run_index}")

    def get_output_path(self, which="merged", **kwargs):
        # try to get the output_filename via 2 successive attempts
        # 1) a getter method if implemented (takes priority)
        # 2) directly via an attribute (fall-back)
        # If none of the two ways work, something is incorrectly implemented,
        # i.e. a developer's problem, not a user problem, and we raise a GateImplementationError
        try:
            output_filename = self.get_output_filename(**kwargs)
        except NotImplementedError:
            try:
                output_filename = getattr(self, "output_filename")
            except AttributeError:
                raise GateImplementationError(
                    f"Unable to get the output_filename "
                    f"in user_output {self.name} "
                    f"of actor {self.belongs_to_actor.name}."
                )
        # 'auto' means that the output_filename is automatically generated.
        if output_filename == "auto":
            output_filename = self._generate_auto_output_filename(**kwargs)
        if output_filename is None or output_filename == "":
            warning(
                f"No output_filename defined for user output '{self.name}' "
                f"of {self.belongs_to_actor.type_name} '{self.belongs_to_actor.name}'. "
                f"Therefore, get_output_path() returns None. "
            )
            return None
        else:
            return self._compose_output_path(which, output_filename)

    def get_output_path_as_string(self, **kwargs):
        return ensure_filename_is_str(self.get_output_path(**kwargs))

    def close(self):
        if self.keep_data_in_memory is False:
            self.data_per_run = {}
            self.merged_data = None
        super().close()

    def get_data(self, **kwargs):
        raise NotImplementedError("This is the base class. ")

    def store_data(self, *args, **kwargs):
        raise NotImplementedError("This is the base class. ")

    def write_data(self, *args, **kwargs):
        raise NotImplementedError("This is the base class. ")

    def write_data_if_requested(self, **kwargs):
        raise NotImplementedError("This is the base class. ")

    def load_data(self, which):
        raise NotImplementedError(
            f"Your are calling this method from the base class {type(self).__name__}, "
            f"but it should be implemented in the specific derived class"
        )


class ActorOutputUsingDataItemContainer(ActorOutputBase):
    # hints for IDE
    merge_data_after_simulation: bool
    keep_data_per_run: bool
    data_item_config: Optional[Box]

    user_info_defaults = {
        "merge_data_after_simulation": (
            True,
            {
                "doc": "In case the simulation has multiple runs, should results from separate runs be merged?"
            },
        ),
        "keep_data_per_run": (
            False,
            {
                "doc": "In case the simulation has multiple runs, should separate results per run be kept?"
            },
        ),
    }

    # this intermediate base class defines a class attribute data_container_class,
    # but leaves it as None. Specific classes need to set it to the correct class or tuple of classes
    data_container_class = None
    _default_interface_class = UserInterfaceToActorOutputUsingDataItemContainer
    _default_data_item_config = None

    @classmethod
    def get_user_info_default_values_interface(cls, item=0, **kwargs):
        defaults = super().get_user_info_default_values_interface(**kwargs)
        for k, v in cls._default_data_item_config[item].items():
            defaults[k] = v
        return defaults

    @classmethod
    def set_user_info_default_values_interface(cls, item=0, **kwargs):
        # pick up the defaults to be stored in the default data item config dictionary
        # and let the base class handle the rest
        known_defaults = list(cls._default_data_item_config[item].keys())
        for k in known_defaults:
            if k in kwargs:
                cls._default_data_item_config[item][k] = kwargs.pop(k)
        super().set_user_info_default_values_interface(**kwargs)

    @classmethod
    def __process_this__(cls):
        super().__process_this__()
        # we need to fill the _default_data_item_config of this class
        # depending on the container classes it handles
        if cls.data_container_class is not None:
            cls._default_data_item_config = {}
            # ask the container class which data item (including aliases and effective names)
            # it handles so that we can corerctly populate the defaults
            for k in cls.data_container_class.__get_data_item_names__():
                if isinstance(k, (int,)):
                    suffix = f"item{k}"
                else:
                    suffix = k
                cls._default_data_item_config[k] = {
                    "output_filename": "auto",
                    "write_to_disk": True,
                    "active": False,
                    "suffix": suffix,
                }
            # if there is only one item, set suffix to None
            # because we do not want to append anything to the output_filename in this case
            # and activate the output
            if len(cls._default_data_item_config) == 1:
                list(cls._default_data_item_config.values())[0]["suffix"] = None
                list(cls._default_data_item_config.values())[0]["active"] = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_item_config = copy.deepcopy(self._default_data_item_config)

    def initialize_cpp_parameters(self):
        # Create structs on C++ side for this actor output
        # This struct is only needed by actors that handle output written in C++.
        # But it does not hurt to populate the info in C++ regardless of the actor
        # The output path can also be (re-)set by the specific actor in
        # StartSimulation or BeginOfRunActionMasterThread, if needed
        items = self._collect_item_identifiers("all")
        for h in items:
            identifier = f"{self.name}_{h}"
            self.belongs_to_actor.AddActorOutputInfo(identifier)
            self.belongs_to_actor.SetWriteToDisk(
                identifier, self.get_write_to_disk(item=h)
            )
            self.belongs_to_actor.SetOutputPath(
                identifier, self.get_output_path_as_string(item=h)
            )

    def to_dictionary(self):
        d = super().to_dictionary()
        d["data_item_config"] = self.data_item_config
        return d

    def from_dictionary(self, d):
        super().from_dictionary(d)
        self.data_item_config = d["data_item_config"]

    def _fatal_unknown_item(self, item):
        fatal(
            f"Unknown item {item}. Known items are {list(self.data_item_config.keys())}."
        )

    # override methods:
    def set_write_to_disk(self, value, item=0):
        items = self._collect_item_identifiers(item)
        for i in items:
            self.data_item_config[i]["write_to_disk"] = bool(value)

    def get_write_to_disk(self, item=0):
        items = self._collect_item_identifiers(item)
        return any([self.data_item_config[k]["write_to_disk"] is True for k in items])

    def set_active(self, value, item=0):
        items = self._collect_item_identifiers(item)
        for i in items:
            self.data_item_config[i]["active"] = bool(value)

    def get_active(self, item=0):
        if item == "any":
            item = "all"
        items = self._collect_item_identifiers(item)
        return any([self.data_item_config[k]["active"] is True for k in items])

    def set_output_filename(self, value, item=0):
        if item == "all":
            for k, v in self.data_item_config.items():
                self.set_output_filename(self._insert_item_suffix(value, k), item=k)
        else:
            try:
                self.data_item_config[item]["output_filename"] = str(value)
            except KeyError:
                self._fatal_unknown_item(item)

    def get_output_filename(self, item=0):
        if item == "all":
            return dict(
                [(k, self.get_output_filename(item=k)) for k in self.data_item_config]
            )
        else:
            try:
                return self.data_item_config[item]["output_filename"]
            except KeyError:
                self._fatal_unknown_item(item)

    def get_item_suffix(self, item=0, **kwargs):
        if item == "all":
            return dict(
                [(k, self.get_item_suffix(item=k)) for k in self.data_item_config]
            )
        else:
            try:
                # FIXME: the .get() method implicitly defines a default value, but it should not. Is this a workaround?
                # return self.data_item_config[item].get("suffix", str(item))
                return self.data_item_config[item]["suffix"]
            except KeyError:
                self._fatal_unknown_item(item)

    def set_item_suffix(self, value, item=0, **kwargs):
        items = self._collect_item_identifiers(item)
        if len(items) > 1:
            fatal("You can set the item suffix only for one item at a time. ")
        self.data_item_config[item]["suffix"] = value

    def _generate_auto_output_filename(self, item=0):
        # try to get a suffix from the data item config dictionary
        # and fall back to the item name (or index) in case no suffix is found
        output_filename = super()._generate_auto_output_filename()
        return self._insert_item_suffix(output_filename, item)

    def _insert_item_suffix(self, output_filename, item):
        suffix = self.data_item_config[item].get("suffix", str(item))
        if suffix is not None:
            output_filename = insert_suffix_before_extension(output_filename, suffix)
        return output_filename

    def _collect_item_identifiers(self, item):
        if item == "all":
            items = list(self.data_item_config.keys())
        elif isinstance(item, (tuple, list)):
            items = item
        else:
            items = [item]
        if not all([i in self.data_item_config for i in items]):
            fatal(
                f"Unknown items. Requested items are: {items}. "
                f"Known items are {list(self.data_item_config.keys())}."
            )
        return items

    def get_output_path(
        self, which="merged", item=0, always_return_dict=False, **kwargs
    ):
        return_dict = {}
        for i in self._collect_item_identifiers(item):
            return_dict[i] = super().get_output_path(which=which, item=i)
        if len(return_dict) > 1 or always_return_dict is True:
            return return_dict
        else:
            return list(return_dict.values())[0]

    def get_data_container(self, which):
        if which == "merged":
            return self.merged_data
        else:
            try:
                run_index = int(which)  # might be a run_index
                if (
                    run_index in self.data_per_run
                    and self.data_per_run[run_index] is not None
                ):
                    return self.data_per_run[run_index]
                else:
                    fatal(f"No data stored for run index {run_index}")
            except ValueError:
                fatal(
                    f"Invalid argument 'which' in get_data_container() method of ActorOutput {self.name}. "
                    f"Allowed values are: 'merged' or a valid run_index. "
                )

    def get_data(self, which="merged", item=0):
        container = self.get_data_container(which)
        if container is None:
            return None
        else:
            return container.get_data(item=item)

    def store_data(self, which, *data):
        """data can be either the user data to be wrapped into a DataContainer class or
        an already wrapped DataContainer class.
        """

        if isinstance(data, self.data_container_class):
            data_container = data
            data_container.belongs_to = self
        else:
            data_container = self.data_container_class(belongs_to=self, data=data)
        # FIXME: use store_data if target data exists, otherwise create new container
        if which == "merged":
            self.merged_data = data_container
        else:
            try:
                run_index = int(which)  # might be a run_index
            except ValueError:
                fatal(
                    f"Invalid argument 'which' in store_data() method of ActorOutput {self.name}. "
                    f"Allowed values are: 'merged' or a valid run_index. "
                )
                run_index = None  # avoid IDE warning
            self.data_per_run[run_index] = data_container

    def store_meta_data(self, which, **meta_data):
        """data can be either the user data to be wrapped into a DataContainer class or
        an already wrapped DataContainer class.
        """

        data = self.get_data_container(which)
        data.update_meta_data(meta_data)

    def load_data(self, which):
        raise NotImplementedError(
            f"Your are calling this method from the base class {type(self).__name__}, "
            f"but it should be implemented in the specific derived class"
        )

    def collect_data(self, which, return_identifier=False):
        if which == "merged":
            data = [self.merged_data]
            identifiers = ["merged"]
        elif which == "all_runs":
            data = list(self.data_per_run.values())
            identifiers = list(self.data_per_run.keys())
        elif which == "all":
            data = list(self.data_per_run.values())
            data.append(self.merged_data)
            identifiers = list(self.data_per_run.keys())
            identifiers.append("merged")
        else:
            try:
                ri = int(which)
            except ValueError:
                fatal(f"Invalid argument which in method collect_images(): {which}")
                ri = None  # avoid IDE warning
            data = [self.data_per_run[ri]]
            identifiers = [ri]
        if return_identifier is True:
            return data, identifiers
        else:
            return data

    def write_data(self, which="all", item="all", **kwargs):
        if which == "all_runs":
            for k in self.data_per_run.keys():
                self.write_data(which=k, item=item, **kwargs)
        elif which == "all":
            self.write_data(which="all_runs", item=item, **kwargs)
            self.write_data(which="merged", item=item, **kwargs)
        else:
            data = self.get_data_container(which)
            if data is not None:
                items = self._collect_item_identifiers(item)
                for i in items:
                    data.write(
                        self.get_output_path(which=which, item=i, **kwargs), item=i
                    )

    def write_data_if_requested(self, which="all", item="all", **kwargs):
        items = [
            i
            for i in self._collect_item_identifiers(item)
            if self.get_write_to_disk(item=i) is True
            and self.get_active(item=i) is True
            # FIXME: the active is True check should not be here. self.write_data() should handle that
        ]
        self.write_data(which=which, item=items)

    def merge_data_from_runs(self):
        self.merged_data = merge_data(list(self.data_per_run.values()))

    def end_of_run(self, run_index):
        if self.merge_data_after_simulation is True:
            self.merged_data.inplace_merge_with(self.data_per_run[run_index])
        if self.keep_data_per_run is False:
            self.data_per_run.pop(run_index)

    def start_of_simulation(self, **kwargs):
        if self.merge_data_after_simulation is True:
            self.merged_data = self.data_container_class(belongs_to=self)

    def end_of_simulation(self, item="all", **kwargs):
        try:
            self.write_data_if_requested(item="all", **kwargs)
        except NotImplementedError:
            raise GateImplementationError(
                "Unable to run end_of_simulation "
                f"in user_output {self.name} of actor {self.belongs_to_actor.name}"
                f"because the class does not implement a write_data_if_requested() "
                f"and/or write_data() method. "
                f"A developer needs to fix this. "
            )


class ActorOutputImage(ActorOutputUsingDataItemContainer):
    _default_interface_class = UserInterfaceToActorOutputImage

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_suffix = "mhd"

    def set_image_properties(self, which, **kwargs):
        for image_data in self.collect_data(which):
            if image_data is not None:
                image_data.set_image_properties(**kwargs)

    def get_image_properties(self, which, item=0):
        if which == "merged":
            if self.merged_data is not None:
                return self.merged_data.get_image_properties()[item]
        else:
            try:
                run_index = int(which)
                try:
                    image_data_container = self.data_per_run[run_index]
                except KeyError:
                    fatal(f"No data found for run index {run_index}.")
                    image_data_container = None  # avoid IDE warning
                if image_data_container is not None:
                    return image_data_container.get_image_properties()[item]
            except ValueError:
                fatal(
                    f"Illegal argument 'which'. Provide a valid run index or the term 'merged'."
                )

    def create_empty_image(self, run_index, size, spacing, origin=None, **kwargs):
        if run_index not in self.data_per_run:
            self.data_per_run[run_index] = self.data_container_class(belongs_to=self)
        self.data_per_run[run_index].create_empty_image(
            size, spacing, origin=origin, **kwargs
        )


# concrete classes usable in Actors:
class ActorOutputSingleImage(ActorOutputImage):
    data_container_class = SingleItkImage


class ActorOutputSingleMeanImage(ActorOutputImage):
    data_container_class = SingleMeanItkImage


class ActorOutputSingleImageWithVariance(ActorOutputImage):
    data_container_class = SingleItkImageWithVariance


class ActorOutputQuotientImage(ActorOutputImage):
    data_container_class = QuotientItkImage


class ActorOutputQuotientMeanImage(ActorOutputImage):
    data_container_class = QuotientMeanItkImage


class ActorOutputRoot(ActorOutputBase):
    # hints for IDE
    output_filename: str
    write_to_disk: bool
    keep_data_in_memory: bool

    user_info_defaults = {
        "output_filename": (
            "auto",
            {
                "doc": "Filename for the data represented by this actor output. "
                "Relative paths and filenames are taken "
                "relative to the global simulation output folder "
                "set via the Simulation.output_dir option. ",
            },
        ),
        "write_to_disk": (
            True,
            {
                "doc": "Should the output be written to disk, or only kept in memory? ",
            },
        ),
        "keep_data_in_memory": (
            False,
            {
                "doc": "Should the data be kept in memory after the end of the simulation? "
                "Otherwise, it is only stored on disk and needs to be re-loaded manually. "
                "Careful: Large data structures like a phase space need a lot of memory. \n"
                "Warning: Feature not supported for ROOT output yet. The options is forced to False. ",
                "override": True,
                "read_only": True,
            },
        ),
    }

    default_suffix = "root"

    @classmethod
    def get_user_info_default_values_interface(cls, **kwargs):
        defaults = super().get_user_info_default_values_interface(**kwargs)
        for k in ["output_filename", "write_to_disk"]:
            defaults[k] = cls.inherited_user_info_defaults[k][0]
        return defaults

    @classmethod
    def set_user_info_default_values_interface(cls, **kwargs):
        for k in ["output_filename", "write_to_disk"]:
            if k in kwargs:
                current_default_tuple = cls.inherited_user_info_defaults[k]
                cls.inherited_user_info_defaults[k] = (
                    kwargs.pop(k),
                    current_default_tuple[1],
                )
        super().set_user_info_default_values_interface(**kwargs)

    def get_output_path(self, *args, **kwargs):
        if "which" in kwargs and kwargs["which"] != "merged":
            self.warn_user(
                "Currently, GATE 10 only stores cumulative ROOT output per simulation ('merged'), "
                "not data per run. Showing you the path to the ROOT file with cumulative data."
            )
        return super().get_output_path(which="merged")

    def initialize(self):
        # Warning, for the moment, MT and root output does not work on windows machine
        if sys.platform.startswith("nt"):
            if g4.IsMultithreadedApplication():
                fatal(
                    f"Sorry Multithreading and Root output does not work (yet) on windows architecture."
                    f"You can run the simulation in single-threaded mode of switch to linux/max."
                )

        # for ROOT output, not output_filename means no output to disk (legacy Gate 9 behavior)
        if self.output_filename == "" or self.output_filename is None:
            self.write_to_disk = False
        self.initialize_cpp_parameters()
        super().initialize()

    def initialize_cpp_parameters(self):
        self.belongs_to_actor.AddActorOutputInfo(self.name)
        self.belongs_to_actor.SetWriteToDisk(self.name, self.write_to_disk)
        if self.output_filename == "" or self.output_filename is None:
            # this test avoid a warning in get_output_path when it is None
            self.belongs_to_actor.SetOutputPath(self.name, "None")
        else:
            self.belongs_to_actor.SetOutputPath(
                self.name, self.get_output_path_as_string()
            )


process_cls(ActorOutputBase)
process_cls(ActorOutputUsingDataItemContainer)
process_cls(ActorOutputImage)
process_cls(ActorOutputSingleImage)
process_cls(ActorOutputSingleMeanImage)
process_cls(ActorOutputSingleImageWithVariance)
process_cls(ActorOutputQuotientImage)
process_cls(ActorOutputQuotientMeanImage)
process_cls(ActorOutputRoot)



### Derenzo phantom with 6 sets of cylinders of different size for each set
### For tubs1 set of cylinders each cylinder is assigned with an activity value
### The same procedure as in tubs1 with the activity can be followed to all phantom sets

import opengate as gate
from scipy.spatial.transform import Rotation as R

# Define the units used in the simulation set-up
cm = gate.g4_units.cm
keV = gate.g4_units.keV
mm = gate.g4_units.mm


def add_derenzo_phantom(sim, name="derenzo"):
    """
    Add a Derenzo phantom with 6 sets of cylinders of different size for each set.
    The phantom is described in https://doi.org/10.3390/diagnostics15111387
    """
    phantom_body = sim.add_volume("Box", name)
    phantom_body.size = [30 * cm, 30 * cm, 5 * cm]
    phantom_body.mother = "world"
    phantom_body.material = "G4_AIR"
    rot = R.from_euler("x", [-90], degrees=True)
    phantom_body.rotation = rot.as_matrix()
    yellow = [1, 1, 0, 0.5]

    tubs_1 = sim.add_volume("Tubs", f"{phantom_body.name}_tubs_1")
    tubs_1.material = "G4_WATER"
    tubs_1.mother = phantom_body.name
    tubs_1.rmin = 0 * mm
    tubs_1.rmax = 14.5 * mm
    tubs_1.dz = 12 * mm
    tubs_1.color = yellow
    tubs_1.translation = [29 * mm, 1.5 * mm, 105 * mm]
    m = R.identity().as_matrix()
    tubs_1.translation = [
        [29 * mm, 105 * mm, 0],
        [-29 * mm, 105 * mm, 0],
        [0 * mm, 54.77 * mm, 0],
    ]

    tubs_2 = sim.add_volume("Tubs", f"{phantom_body.name}_tubs_2")
    tubs_2.material = "G4_WATER"
    tubs_2.mother = phantom_body.name
    tubs_2.rmin = 0 * mm
    tubs_2.rmax = 9.3 * mm
    tubs_2.dz = 12 * mm
    tubs_2.color = yellow
    tubs_2.translation = [29 * mm, 1.5 * mm, 105 * mm]
    m = R.identity().as_matrix()
    tubs_2.translation = [
        [66 * mm, 25 * mm, 0],
        [66 * mm, 89.44 * mm, 0],
        [28.8 * mm, 25 * mm, 0],
        [103.2 * mm, 25 * mm, 0],
        [84.5 * mm, 57.21 * mm, 0],
        [47.5 * mm, 57.21 * mm, 0],
    ]

    tubs_3 = sim.add_volume("Tubs", f"{phantom_body.name}_tubs_3")
    tubs_3.material = "G4_WATER"
    tubs_3.mother = phantom_body.name
    tubs_3.rmin = 0 * mm
    tubs_3.rmax = 7.85 * mm
    tubs_3.dz = 12 * mm
    tubs_3.color = yellow
    tubs_3.translation = [29 * mm, 1.5 * mm, 105 * mm]
    m = R.identity().as_matrix()
    tubs_3.translation = [
        [66 * mm, -45.19 * mm, 0],
        [97.4 * mm, -45.19 * mm, 0],
        [34.6 * mm, -45.19 * mm, 0],
        [18.9 * mm, -18 * mm, 0],
        [50.3 * mm, -18 * mm, 0],
        [81.7 * mm, -18 * mm, 0],
        [113.1 * mm, -18 * mm, 0],
        [50.3 * mm, -72.38 * mm, 0],
        [81.7 * mm, -72.38 * mm, 0],
        [66 * mm, -99.57 * mm, 0],
    ]

    tubs_4 = sim.add_volume("Tubs", f"{phantom_body.name}_tubs_4")
    tubs_4.material = "G4_WATER"
    tubs_4.mother = phantom_body.name
    tubs_4.rmin = 0 * mm
    tubs_4.rmax = 6.5 * mm
    tubs_4.dz = 12 * mm
    tubs_4.color = yellow
    tubs_4.translation = [29 * mm, 1.5 * mm, 105 * mm]
    m = R.identity().as_matrix()
    tubs_4.translation = [
        [0 * mm, -91.69 * mm, 0],
        [0 * mm, -46.67 * mm, 0],
        [26 * mm, -91.69 * mm, 0],
        [-26 * mm, -91.69 * mm, 0],
        [-13 * mm, -69.18 * mm, 0],
        [13 * mm, -69.18 * mm, 0],
        [13 * mm, -114.2 * mm, 0],
        [-13 * mm, -114.2 * mm, 0],
        [-39 * mm, -114.2 * mm, 0],
        [39 * mm, -114.2 * mm, 0],
    ]

    tubs_5 = sim.add_volume("Tubs", f"{phantom_body.name}_tubs_5")
    tubs_5.material = "G4_WATER"
    tubs_5.mother = phantom_body.name
    tubs_5.rmin = 0 * mm
    tubs_5.rmax = 5.75 * mm
    tubs_5.dz = 12 * mm
    tubs_5.color = yellow
    tubs_5.translation = [29 * mm, 1.5 * mm, 105 * mm]
    m = R.identity().as_matrix()
    tubs_5.translation = [
        [-66 * mm, -42.92 * mm, 0],
        [-89 * mm, -42.92 * mm, 0],
        [-43 * mm, -42.92 * mm, 0],
        [-66 * mm, -82.76 * mm, 0],
        [-54.5 * mm, -62.84 * mm, 0],
        [-77.5 * mm, -62.84 * mm, 0],
        [-54.5 * mm, -23 * mm, 0],
        [-77.5 * mm, -23 * mm, 0],
        [-31.5 * mm, -23 * mm, 0],
        [-100.5 * mm, -23 * mm, 0],
    ]

    tubs_6 = sim.add_volume("Tubs", f"{phantom_body.name}_tubs_6")
    tubs_6.material = "G4_WATER"
    tubs_6.mother = phantom_body.name
    tubs_6.rmin = 0 * mm
    tubs_6.rmax = 5 * mm
    tubs_6.dz = 12 * mm
    tubs_6.color = yellow
    tubs_6.translation = [29 * mm, 1.5 * mm, 105 * mm]
    m = R.identity().as_matrix()
    tubs_6.translation = [
        [-66 * mm, 17 * mm, 0],
        [-46 * mm, 17 * mm, 0],
        [-26 * mm, 17 * mm, 0],
        [-86 * mm, 17 * mm, 0],
        [-106 * mm, 17 * mm, 0],
        [-56 * mm, 34.32 * mm, 0],
        [-36 * mm, 34.32 * mm, 0],
        [-76 * mm, 34.32 * mm, 0],
        [-96 * mm, 34.32 * mm, 0],
        [-66 * mm, 51.64 * mm, 0],
        [-46 * mm, 51.64 * mm, 0],
        [-86 * mm, 51.64 * mm, 0],
        [-76 * mm, 68.96 * mm, 0],
        [-56 * mm, 68.96 * mm, 0],
        [-66 * mm, 86.28 * mm, 0],
    ]
    return phantom_body


def add_sources(sim, derenzo_phantom, activity_Bq_mL):
    """
    The source is attached to the tubs volumes of the derenzo,
    it means its coordinate system is the same
    activity_Bq_mL should contain the activity concentration for each set of tubs (6)
    """

    sources = []
    for nb_tub in range(1, 7):
        tubs = sim.volume_manager.volumes[f"{derenzo_phantom.name}_tubs_{nb_tub}"]
        for i in range(len(tubs.translation)):
            tub_name = tubs.get_repetition_name_from_index(i)
            s = tubs.solid_info
            source = sim.add_source(
                "GenericSource", f"{derenzo_phantom.name}_tubs_{nb_tub}_source_{i}"
            )
            source.attached_to = tub_name
            source.particle = "gamma"
            source.energy.mono = 140 * keV
            source.activity = activity_Bq_mL[nb_tub - 1] * s.cubic_volume
            source.position.type = "cylinder"
            source.position.radius = tubs.rmax
            source.position.dz = tubs.dz
            source.direction.type = "iso"
            sources.append(source)

    return sources



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test023")

    # create the simulation
    sim = gate.Simulation()

    # main options
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.random_seed = 321456987
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm
    mm = gate.g4_units.mm
    sec = gate.g4_units.second
    keV = gate.g4_units.keV

    #  change world size
    sim.world.size = [1 * m, 1 * m, 1 * m]
    sim.world.material = "G4_AIR"

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.type = "gauss"
    source.energy.mono = 1 * MeV
    source.energy.sigma_gauss = 1 * MeV
    source.particle = "gamma"
    source.position.type = "sphere"
    source.position.radius = 1 * cm
    source.direction.type = "iso"
    source.activity = 1000 * Bq

    print(source)

    # plane
    plane1a = sim.add_volume("Box", "plane1a")
    plane1a.size = [1 * m, 1 * m, 1 * nm]
    plane1a.translation = [0, 0, 1 * cm]
    plane1a.material = "G4_AIR"
    plane1a.color = [1, 0, 0, 1]  # red

    plane1b = sim.add_volume("Box", "plane1b")
    plane1b.size = [1 * m, 1 * m, 1 * nm]
    plane1b.translation = [0, 0, 2 * cm]
    plane1b.material = "G4_AIR"
    plane1b.color = [1, 0, 0, 1]  # red

    plane2a = sim.add_volume("Box", "plane2a")
    plane2a.size = [1 * m, 1 * m, 1 * nm]
    plane2a.translation = [0, 0, -1 * cm]
    plane2a.material = "G4_AIR"
    plane2a.color = [0, 1, 0, 1]  # green

    plane2b = sim.add_volume("Box", "plane2b")
    plane2b.size = [1 * m, 1 * m, 1 * nm]
    plane2b.translation = [0, 0, -2 * cm]
    plane2b.material = "G4_AIR"
    plane2b.color = [0, 1, 0, 1]  # green

    # kill according to time
    ka = sim.add_actor("KillActor", "kill_actor1")
    ka.attached_to = plane1a.name
    att_filter = sim.add_filter("ThresholdAttributeFilter", "time_filter")
    # we don't kill the particle within the time range, so
    # we discard the kill filter when the time is in the correct range
    att_filter.attribute = "GlobalTime"
    att_filter.value_min = 20 * sec
    att_filter.value_max = 70 * sec
    att_filter.policy = "reject"
    print(att_filter)
    ka.filters.append(att_filter)

    # kill according to energy
    ka = sim.add_actor("KillActor", "kill_actor2")
    ka.attached_to = plane2a.name
    att_filter = sim.add_filter("ThresholdAttributeFilter", "ene_filter")
    att_filter.attribute = "KineticEnergy"
    att_filter.value_min = 300 * keV
    att_filter.value_max = 1200 * keV
    att_filter.policy = "accept"
    ka.filters.append(att_filter)

    # stats
    stat = sim.add_actor("SimulationStatisticsActor", "stats")
    stat.track_types_flag = True

    # phsp
    phsp1 = sim.add_actor("PhaseSpaceActor", "phsp1")
    # warning, if the plane is plane1a the particle may be marked as "killed"
    # but will still be in the phsp
    phsp1.attached_to = plane1b.name
    phsp1.attributes = ["GlobalTime", "KineticEnergy"]
    phsp1.output_filename = f"test023_filters_attribute.root"
    phsp1.priority = ka.priority + 10

    # phsp
    phsp2 = sim.add_actor("PhaseSpaceActor", "phsp2")
    phsp2.attached_to = plane2b.name
    phsp2.attributes = ["GlobalTime", "KineticEnergy"]
    phsp2.output_filename = f"test023_filters_attribute.root"

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"

    # start simulation
    duration = 100 * sec
    sim.run_timing_intervals = [[0, duration]]
    sim.run()

    # print results at the end
    print(stat)
    # reference :
    # stat.write(paths.output_ref / "test023_att_stats.txt")

    # tests
    stats_ref = utility.read_stat_file(paths.output_ref / "test023_att_stats.txt")
    is_ok = utility.assert_stats(stat, stats_ref, 0.01)

    # image root files (no comparison here, just for plot)
    print()
    print("WARNING: no comparison here, only for plot")
    utility.compare_root3(
        phsp1.get_output_path(),
        phsp1.get_output_path(),
        "phsp1",
        "phsp2",
        keys1=phsp1.attributes,
        keys2=phsp1.attributes,
        tols=[1e20] * 2,
        scalings1=[1e-9, 1.0],
        scalings2=[1e-9, 1.0],
        img=paths.output / "test023_att.png",
        hits_tol=1e20,
    )

    # compare root files (no comparison here)
    print()
    is_ok = (
        utility.compare_root3(
            paths.output_ref / f"test023_filters_attribute.root",
            phsp1.get_output_path(),
            "phsp1",
            "phsp1",
            keys1=phsp1.attributes,
            keys2=phsp1.attributes,
            tols=[0.4, 0.02],
            scalings1=[1e-9, 1.0],
            scalings2=[1e-9, 1.0],
            img=paths.output / "test023_att_phsp1.png",
            hits_tol=2,
        )
        and is_ok
    )
    print()
    is_ok = (
        utility.compare_root3(
            paths.output_ref / f"test023_filters_attribute.root",
            phsp1.get_output_path(),
            "phsp2",
            "phsp2",
            keys1=phsp1.attributes,
            keys2=phsp1.attributes,
            tols=[0.6, 0.02],
            scalings1=[1e-9, 1.0],
            scalings2=[1e-9, 1.0],
            img=paths.output / "test023_att_phsp2.png",
            hits_tol=2,
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
from opengate.tests import utility

# ABOUT THIS TEST:
# This test shows how a user can implement a customized version of a geometry changer
# and add it manually to the simulation
# This test should perform exactly the same simulation as test030_dose_motion_dynamic_param.py,
# but print out extra info in-between the runs

# See also test071_custom_geometry_changer
# Warning: This is intended for advanced GATE users


def print_face():
    print("*******************")
    print("**** ^^^   ^^^ ****")
    print("****  O     O  ****")
    print("****     L     ****")
    print("****  \_____/  ****")
    print("****     V     ****")
    print("*******************")


class MyCustomTranslationChanger(gate.actors.dynamicactors.VolumeTranslationChanger):
    # The VolumeTranslationChanger has a user_info called 'translations' which holds the list of translation vectors
    # you can access it via self.translations

    # implement an 'apply_change() method
    def apply_change(self, run_id):
        # call the apply_change() method from the superclass
        # which will actually apply the translation
        super().apply_change(run_id)
        # now do your own stuff
        print_face()
        print(
            f"This is MyCustomTranslationChanger working on volume {self.attached_to_volume.name} in run {run_id}."
        )
        print(f"Current translation: {self.translations[run_id]}")
        run_interval = self.volume_manager.simulation.run_timing_intervals[run_id]
        print(
            f"This run interval is from {run_interval[0] / gate.g4_units.s} s "
            f"to {run_interval[1] / gate.g4_units.s} s"
        )
        print()


class MyCustomRotationChanger(gate.actors.dynamicactors.VolumeRotationChanger):
    # The VolumeRotationChanger has a user_info called 'rotations' which holds the list of rotation matrices
    # you can access it via self.rotations

    # implement an 'apply_change() method
    def apply_change(self, run_id):
        # call the apply_change() method from the superclass
        # which will actually apply the translation
        super().apply_change(run_id)
        # now do your own stuff
        print_face()
        print(
            f"This is MyCustomRotationChanger working on volume {self.attached_to_volume.name} in run {run_id}."
        )
        print(f"Current rotations: {self.rotations[run_id]}")
        run_interval = self.volume_manager.simulation.run_timing_intervals[run_id]
        print(
            f"This run interval is from {run_interval[0] / gate.g4_units.s} s "
            f"to {run_interval[1] / gate.g4_units.s} s"
        )
        print()


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test029_volume_time_rotation", "test030"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.random_seed = 983456
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    um = gate.g4_units.um
    nm = gate.g4_units.nm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    sec = gate.g4_units.second

    #  change world size
    sim.world.size = [1 * m, 1 * m, 1 * m]

    # add a simple fake volume to test hierarchy
    # translation and rotation like in the Gate macro
    fake = sim.add_volume("Box", "fake")
    fake.size = [40 * cm, 40 * cm, 40 * cm]
    fake.translation = [1 * cm, 2 * cm, 3 * cm]
    fake.material = "G4_AIR"
    fake.color = [1, 0, 1, 1]

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.mother = "fake"
    waterbox.size = [20 * cm, 20 * cm, 20 * cm]
    waterbox.translation = [-3 * cm, -2 * cm, -1 * cm]
    waterbox.rotation = Rotation.from_euler("y", -20, degrees=True).as_matrix()
    waterbox.material = "G4_WATER"
    waterbox.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.set_production_cut("world", "all", 700 * um)

    # default source for tests
    # the source is fixed at the center, only the volume will move
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 150 * MeV
    source.particle = "proton"
    source.position.type = "disc"
    source.position.radius = 5 * mm
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 30000 * Bq

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = "test030-custdyn-edep.mhd"
    dose.attached_to = "waterbox"
    dose.size = [99, 99, 99]
    mm = gate.g4_units.mm
    dose.spacing = [2 * mm, 2 * mm, 2 * mm]
    dose.translation = [2 * mm, 3 * mm, -2 * mm]
    dose.edep_uncertainty.active = True

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")

    # created the translations and rotations which describe the motion of the volume 'fake'
    n = 3
    interval_length = 1 * sec / n
    sim.run_timing_intervals = [
        (i * interval_length, (i + 1) * interval_length) for i in range(n)
    ]
    gantry_angles_deg = [i * 20 for i in range(n)]
    # use a helper function
    (
        dynamic_translations,
        dynamic_rotations,
    ) = gate.geometry.utility.get_transform_orbiting(
        initial_position=fake.translation, axis="Y", angle_deg=gantry_angles_deg
    )

    # create the changers manually
    translation_changer = MyCustomTranslationChanger(name="translation_changer")
    rotation_changer = MyCustomRotationChanger(name="rotation_changer")
    # set the translations and rotations
    translation_changer.translations = dynamic_translations
    translation_changer.attached_to = fake
    rotation_changer.rotations = dynamic_rotations
    rotation_changer.attached_to = fake

    # add a DynamicGeometryActor to the simulation
    dyn_geo_actor = sim.add_actor("DynamicGeometryActor", name="dyn_geo_actor")
    # ... and add the changers to the actor
    dyn_geo_actor.geometry_changers.append(translation_changer)
    dyn_geo_actor.geometry_changers.append(rotation_changer)

    # start simulation
    sim.run()

    # print results at the end
    print(stats)

    # tests
    stats_ref = utility.read_stat_file(paths.output_ref / "stats030.txt")
    is_ok = utility.assert_stats(stats, stats_ref, 0.11)

    print()
    gate.exception.warning("Difference for EDEP")
    is_ok = (
        utility.assert_images(
            paths.output_ref / "test030-edep.mhd",
            dose.edep.get_output_path(),
            stats,
            tolerance=30,
            sum_tolerance=6,
            ignore_value_data2=0,
        )
        and is_ok
    )

    print("\nDifference for uncertainty")
    is_ok = (
        utility.assert_images(
            paths.output_ref / "test030-edep_uncertainty.mhd",
            dose.edep_uncertainty.get_output_path(),
            stats,
            tolerance=15,
            ignore_value_data2=0,
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import uproot
import numpy as np
from opengate.tests import utility


def validation_test(n, n_measured):
    n_part_minus = n - 3 * np.sqrt(n)
    n_part_max = n + 3 * np.sqrt(n)
    if n_part_minus <= n_measured <= n_part_max:
        return True
    else:
        return False


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "test066_PhsSource_Activity", output_folder="test066"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    # sim.running_verbose_level = gate.EVENT
    sim.number_of_threads = 1
    sim.random_seed = 987654321
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    km = gate.g4_units.km
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    nm = gate.g4_units.nm
    Bq = gate.g4_units.Bq
    MeV = gate.g4_units.MeV
    keV = gate.g4_units.keV
    gcm3 = gate.g4_units.g / gate.g4_units.cm3

    #  adapt world size
    world = sim.world
    world.size = [1 * m, 1 * m, 2 * m]
    world.material = "G4_Galactic"

    # source_1
    nb_part = 1000
    plane_1 = sim.add_volume("Box", "plan_1")
    plane_1.material = "G4_Galactic"
    plane_1.mother = world.name
    plane_1.size = [1 * m, 1 * m, 1 * nm]
    plane_1.color = [0.8, 0.2, 0.1, 1]

    source_1 = sim.add_source("PhaseSpaceSource", "phsp_source_global_1")
    source_1.attached_to = plane_1.name
    source_1.phsp_file = paths.output_ref.parent / "test019" / "test019_hits.root"
    source_1.position_key = "PrePosition"
    source_1.direction_key = "PreDirection"
    source_1.weight_key = "Weight"
    source_1.global_flag = False
    source_1.particle = "gamma"
    source_1.batch_size = 100
    source_1.translate_position = True
    source_1.position.translation = [0, 0, 300]
    source_1.activity = nb_part * Bq

    # add phase space plan 1

    phsp_plane_1 = sim.add_volume("Box", "phase_space_plane_1")
    phsp_plane_1.mother = world.name
    phsp_plane_1.material = "G4_Galactic"
    phsp_plane_1.size = [1 * m, 1 * m, 1 * nm]
    phsp_plane_1.translation = [0, 0, -1000 * nm]
    phsp_plane_1.color = [1, 0, 0, 1]  # red

    phsp_actor = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    phsp_actor.attached_to = phsp_plane_1
    phsp_actor.attributes = [
        "KineticEnergy",
    ]

    phsp_actor.output_filename = "test066_output_data.root"

    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.gamma = 1 * mm
    sim.physics_manager.global_production_cuts.electron = 1 * mm
    sim.physics_manager.global_production_cuts.positron = 1 * mm
    #
    # # go !
    sim.run()

    #
    # # print results
    print(stats)
    #
    f_phsp = uproot.open(phsp_actor.get_output_path())
    arr = f_phsp["PhaseSpace"].arrays()
    print("Number of detected events :", len(arr))
    print(
        "Number of expected events :",
        nb_part,
        "+- 3*" + str(int(np.round(np.sqrt(nb_part)))),
    )

    is_ok = validation_test(nb_part, len(arr))
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test014_engine_helpers as test014
import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    sim = gate.Simulation()
    test014.define_simulation(sim)

    # go
    sim.run(start_new_process=True)

    # get output
    is_ok = test014.test_output(sim)

    utility.test_ok(is_ok)



from ..exception import fatal, warning
from ..utility import g4_units, LazyModuleLoader
from .generic import GenericSource
import opengate_core as g4

from box import Box
import jsonpickle
import numpy as np
import pathlib
import urllib
import copy
import os
import inspect
from ..base import process_cls

# the following packages seems to take a bit of time to load
rd = LazyModuleLoader("radioactivedecay")
pandas = LazyModuleLoader("pandas")


class PhotonFromIonDecaySource(GenericSource):
    """
    Manage a set of GenericSource sub_sources, one for each nuclide gamma lines, for all daughters of the given ion.
    Each sub_sources will have:
    - activity managed by a TAC, corresponding to the Bateman equation during the time range
    - spectrum energy line for isomeric transition
    - spectrum energy line for atomic relaxation (fluo)
    - particle forced to gammas
    """

    # hints for IDE
    verbose: bool
    tac_bins: float
    dump_log: str
    atomic_relaxation_flag: bool
    isomeric_transition_flag: bool

    user_info_defaults = {
        "verbose": (False, {"doc": "Verbose for debug"}),
        "tac_bins": (200, {"doc": "Number of bins for the TAC"}),
        "dump_log": (None, {"doc": "write log in the given filename"}),
        "atomic_relaxation_flag": (
            True,
            {"doc": "Consider gammas from atomic relaxation"},
        ),
        "isomeric_transition_flag": (
            True,
            {"doc": "Consider gammas from isomeric transition"},
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        # list of all sub_sources
        self.sub_sources = []
        # False if this is the main first source
        self.is_a_sub_source = False

        # used during sub_sources creation
        self.tac_from_decay_parameters = None
        # used during sub_sources creation
        self.daughters = None

        # debug
        self.log = ""
        self.debug_first_daughter_only = False

    def __initcpp__(self):
        g4.GateGenericSource.__init__(self)

    def initialize(self, run_timing_intervals):
        if not self.is_a_sub_source:
            # this is called only one single time to build the list of sub_sources
            if (g4.IsMultithreadedApplication() and g4.G4GetThreadId() == -1) or (
                not g4.IsMultithreadedApplication()
            ):
                phid_build_all_sub_sources(self)

        # super class conventional initialization
        GenericSource.initialize(self, run_timing_intervals)

        # init timing intervals for all ui
        self.initialize_start_end_time(run_timing_intervals)
        for sub_source in self.sub_sources:
            sub_source.start_time = self.start_time
            sub_source.end_time = self.end_time

        # this will initialize and set user_info to the cpp side
        # for all sub_sources
        for sub_source in self.sub_sources:
            # get tac from decay
            p = Box(sub_source.tac_from_decay_parameters)
            sub_source.tac_times, sub_source.tac_activities = get_tac_from_decay(
                p.ion_name,
                p.daughter,
                sub_source.activity,
                sub_source.start_time,
                sub_source.end_time,
                p.bins,
            )
            # update the tac
            update_sub_source_tac_activity(sub_source)

            # check
            self.check_ui_activity(sub_source)
            self.check_confine(sub_source)

            # final initialize to cpp side
            sub_source.InitializeUserInfo(sub_source.user_info)

        # dump log
        if self.user_info.dump_log is not None:
            with open(self.user_info.dump_log, "w") as outfile:
                outfile.write(self.log)

    def add_to_source_manager(self, source_manager):
        for g4_source in self.sub_sources:
            source_manager.AddSource(g4_source)


def update_sub_source_tac_activity(sub_source):
    if sub_source.tac_times is None and sub_source.tac_activities is None:
        return
    n = len(sub_source.tac_times)
    if n != len(sub_source.tac_activities):
        fatal(
            f"option tac_activities must have the same size than tac_times in source '{sub_source.name}'"
        )

    # scale the activity if energy_spectrum is given (because total may not be 100%)
    total = sum(sub_source.energy.spectrum_weights)
    sec = g4_units.s
    Bq = g4_units.Bq

    # it is important to set the starting time for this source as the tac
    # may start later than the simulation timing
    i = 0
    while i < len(sub_source.tac_activities) and sub_source.tac_activities[i] <= 0:
        i += 1
    if i >= len(sub_source.tac_activities):
        # gate.warning(f"Source '{sub_source.name}' TAC with zero activity.")
        sub_source.start_time = sub_source.end_time + 1 * sec
    else:
        sub_source.start_time = sub_source.tac_times[i]
        # IMPORTANT : activities must be x by total here
        # (not before, because it can be called several times in MT mode)
        sub_source.SetTAC(
            sub_source.tac_times, np.array(sub_source.tac_activities) * total
        )

    if sub_source.verbose:
        print(
            f"GammaFromIon source {sub_source.name}    total = {total * 100:8.2f}%   "
            f" gammas lines = {len(sub_source.energy.spectrum_weights)}   "
            f" total activity = {sum(sub_source.tac_activities) / Bq:10.3f}"
            f" first activity = {sub_source.tac_activities[0] / Bq:5.2f}"
            f" last activity = {sub_source.tac_activities[-1] / Bq:5.2f}"
        )


def print_phid_info(rad_name, br=1.0, tab=""):
    nuclide = get_nuclide_from_name(rad_name)
    print(
        f"{tab}{nuclide.nuclide}    Z={nuclide.Z} A={nuclide.A}     "
        f"HL={nuclide.half_life('readable')} ({nuclide.half_life('s'):.1f} s)"
        f"     BF={br}"
    )
    progeny = nuclide.progeny()
    brs = nuclide.branching_fractions()
    t = tab + "  "
    for p, b in zip(progeny, brs):
        print_phid_info(p, b, t)
    return nuclide


def get_nuclide_from_name(rad_name):
    try:
        return rd.Nuclide(rad_name)
    except:
        fatal(f"Cannot find nuclide named {rad_name}, try something like 225Ac.")


def get_nuclide_and_direct_progeny(z, a):
    a = int(a)
    z = int(z)
    id = int(f"{z:3}{a:3}0000")
    nuclide = rd.Nuclide(id)
    p = nuclide.progeny()
    return nuclide, p


def get_nuclide_progeny(nuclide, intensity=1.0, parent=None):
    # insert current nuclide
    p = []
    if parent is None:
        a = Box()
        a.nuclide = nuclide
        a.hl = a.nuclide.half_life()
        a.parent = [None]
        a.intensity = intensity
        p.append(a)
    # start a list of daughters
    daughters = nuclide.progeny()
    branching_fractions = nuclide.branching_fractions()
    # loop recursively
    # the intensity is the branching fraction x the current intensity
    # if the rad is already in the list, we add the intensity
    nuc_to_add = []
    i = 0
    for d, br in zip(daughters, branching_fractions):
        a = Box()
        try:
            a.nuclide = rd.Nuclide(d)
        except:
            warning(f"Unknown nuclide {d}, ignoring ...")
            continue
        a.hl = a.nuclide.half_life()
        a.parent = [nuclide]
        a.intensity = intensity * br
        p.append(a)
        aa = get_nuclide_progeny(a.nuclide, intensity=a.intensity, parent=nuclide)
        nuc_to_add += aa
        i = i + 1

    # the daughter's daughters are added after the loop to keep the order
    # also : merge parents
    for aa in nuc_to_add:
        found = next(
            (item for item in p if item.nuclide.nuclide == aa.nuclide.nuclide),
            None,
        )
        if found:
            found.intensity += aa.intensity
            found.parent += aa.parent
            # remove duplicate
            found.parent = list(set(found.parent))
        else:
            p.append(aa)
    return p


def atomic_relaxation_load(nuclide, load_type="local"):
    ene_ar, w_ar = None, None
    if load_type == "local":
        ene_ar, w_ar = atomic_relaxation_load_from_file(nuclide.nuclide)
    elif load_type == "iaea":
        filename = atomic_relaxation_get_filename(nuclide.nuclide)
        warning(
            f"Load data for {nuclide.nuclide} from IAEA website and store in : {filename}"
        )
        name = nuclide.nuclide[: nuclide.nuclide.index("-")]
        df = atomic_relaxation_load_from_iaea_website(nuclide.A, name)
        atomic_relaxation_store_to_file(nuclide.nuclide, df)
        ene_ar, w_ar = atomic_relaxation_load_from_file(nuclide.nuclide)
    else:
        df = atomic_relaxation_load_from_data_file(nuclide, load_type)
        atomic_relaxation_store_to_file(nuclide.nuclide, df)
        ene_ar, w_ar = atomic_relaxation_load_from_file(nuclide.nuclide)
    return ene_ar, w_ar


def atomic_relaxation_load_from_data_file(nuclide, filename):
    # get info
    A = nuclide.A
    Z = nuclide.Z
    N = A - Z

    # read data file
    df = pandas.read_csv(filename, header=0, dtype=str, low_memory=False)
    c = ["z_parent", "n_parent", "z_daughter", "n_daughter"]
    df[c] = df[c].astype(int)

    # Filter lines where the first value is Z and the second value is N
    df = df[(df["z_parent"] == Z) & (df["n_parent"] == N)]

    # convert when the intensity is an interval, use the mean
    def convert_to_mean(interval):
        if "-" in interval:
            # Split the interval string and convert values to float
            start, end = map(float, interval.split(" - "))
            # Calculate the mean value
            mean_value = (start + end) / 2
            return mean_value
        else:
            # Return the original value if it's not a string
            return interval

    df["intensity_100_dec_of_parent"] = df["intensity_100_dec_of_parent"].apply(
        convert_to_mean
    )

    # Remove Lx L KA KB and Kx (that are duplicated lines)
    df = df[~df["shell"].str.match(r"^L\d+$")]
    df = df[~df["shell"].str.match(r"^L$")]
    df = df[~df["shell"].str.match(r"^KA$")]
    df = df[~df["shell"].str.match(r"^KB$")]
    df = df[~df["shell"].str.match(r"^K\d+$")]

    # rename columns and convert to float
    df = df.rename(
        columns={
            "energy": "energy_old",
            "energy_num": "energy",
            "intensity_100_dec_of_parent": "intensity",
        }
    )
    df["intensity"] = df["intensity"].astype(float)
    df["energy"] = df["energy"].astype(float)

    # group by same intensity
    df = df.groupby("energy")["intensity"].sum()
    df = df.reset_index(name="intensity")
    print(f"Total number of lines: {len(df)}")

    return df


def atomic_relaxation_load_from_iaea_website(a, rad_name):
    # https://nds.iaea.org/relnsd/vcharthtml/VChartHTML.html
    livechart = "https://nds.iaea.org/relnsd/v1/data?"
    nuclide_name = f"{a}{rad_name}"
    url = livechart + f"fields=decay_rads&nuclides={nuclide_name}&rad_types=x"
    print(url)
    try:
        df = lc_read_csv(url)
        print(df)
    except:
        raise Exception(
            f"Cannot get data for atomic relaxation of {rad_name} with this url : {url}"
        )
    if "intensity" not in df:
        # when there is no xray
        return None
    return df


def atomic_relaxation_get_filename(nuclide_name):
    gate_module = inspect.getfile(inspect.importlib.import_module("opengate"))
    folder = pathlib.Path(os.path.dirname(gate_module)) / "data" / "atomic_relaxation"
    filename = folder / f"{nuclide_name.lower()}.txt"
    return filename


def atomic_relaxation_load_from_file(nuclide_name, filename=None):
    nuclide_name = nuclide_name.lower()
    if filename is None:
        filename = atomic_relaxation_get_filename(nuclide_name)
    try:
        df = pandas.read_csv(filename)
    except pandas.errors.EmptyDataError:
        return [], []
    except FileNotFoundError:
        raise Exception(
            f"During 'load_ion_gamma_atomic_relaxation_nds_iaea' cannot read file"
            f" {nuclide_name}.txt in {filename}"
        )
    try:
        ene, w = atomic_relaxation_get_ene_weights_from_df(df)
    except:
        return [], []
    return ene, w


def atomic_relaxation_store_to_file(nuclide_name, df, filename=None):
    nuclide_name = nuclide_name.lower()
    if filename is None:
        filename = atomic_relaxation_get_filename(nuclide_name)
    warning(f"Store atomic relaxation data for {nuclide_name} in: {filename}")
    if df is not None:
        df.to_csv(filename, index=False)
    else:
        f = open(filename, "w")
        f.close()


def atomic_relaxation_get_ene_weights_from_df(df):
    if df is None:
        return np.array([]), np.array([])
    # remove blanks (unknown intensities)
    df = df[pandas.to_numeric(df["intensity"], errors="coerce").notna()]
    # convert to numeric. Note how one can specify the field by attribute or by string
    keV = g4_units.keV
    df.energy = df["energy"].astype(float)
    df.intensity = df["intensity"].astype(float)
    return df.energy.to_numpy() * keV, df.intensity.to_numpy() / 100


def atomic_relaxation_load_all_gammas(nuclide):
    daughters = get_nuclide_progeny(nuclide)
    results = []
    for d in daughters:
        ene_ar, w_ar = atomic_relaxation_load(d.nuclide)
        for e, w in zip(ene_ar, w_ar):
            results.append({"energy": e, "intensity": w, "type": "ar", "nuclide": d})
    return results


def lc_read_csv(url):
    req = urllib.request.Request(url)
    req.add_header(
        "User-Agent",
        "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:77.0) Gecko/20100101 Firefox/77.0",
    )
    return pandas.read_csv(urllib.request.urlopen(req))


def isomeric_transition_get_filename(nuclide_name):
    gate_module = inspect.getfile(inspect.importlib.import_module("opengate"))
    folder = pathlib.Path(os.path.dirname(gate_module)) / "data" / "isomeric_transition"
    filename = folder / f"{nuclide_name.lower()}.txt"
    return filename


def isomeric_transition_load(nuclide, filename=None, half_life=None):
    if filename is None:
        filename = isomeric_transition_get_filename(nuclide.nuclide)
    if half_life is None:
        sec = g4_units.s
        half_life = nuclide.half_life("s") * sec
    try:
        ene, w = isomeric_transition_load_from_df_file(
            nuclide.nuclide, half_life=half_life, filename=filename
        )
        return ene, w
    except Exception:
        name = nuclide.nuclide[: nuclide.nuclide.index("-")]
        df = isomeric_transition_load_from_iaea_website(nuclide.A, name)
        isomeric_transition_store_df_to_file(nuclide.nuclide, df, filename)
        ene, w = isomeric_transition_get_ene_weights_from_df(df, half_life=half_life)
        return np.array(ene), np.array(w)


def isomeric_transition_store(nuclide_name, data_to_save, filename):
    jsonpickle.handlers.registry.register(np.ndarray, NumpyArrayHandler)
    frozen = jsonpickle.encode(data_to_save, indent=2)
    warning(f"Store isomeric transition data for {nuclide_name} in: {filename}")
    if filename is None:
        filename = isomeric_transition_get_filename(nuclide_name)
    with open(filename, "w") as outfile:
        outfile.write(frozen)


def isomeric_transition_store_df_to_file(nuclide_name, df, filename=None):
    nuclide_name = nuclide_name.lower()
    if filename is None:
        filename = isomeric_transition_get_filename(nuclide_name)
    if df is not None:
        df.to_csv(filename, index=False)
    else:
        f = open(filename, "w")
        f.close()


def isomeric_transition_load_from_df_file(nuclide_name, half_life, filename=None):
    nuclide_name = nuclide_name.lower()
    if filename is None:
        filename = isomeric_transition_get_filename(nuclide_name)
    try:
        df = pandas.read_csv(filename)
    except pandas.errors.EmptyDataError:
        return [], []
    except FileNotFoundError:
        raise Exception(
            f"During 'isomeric_transition_load_from_df_file' cannot read file"
            f" {nuclide_name}.txt in {filename}"
        )
    ene, w = isomeric_transition_get_ene_weights_from_df(df, half_life=half_life)
    return ene, w


def isomeric_transition_get_ene_weights_from_df(df, half_life):
    if df is None:
        return np.array([]), np.array([])
    # remove blanks (unknown intensities)
    df = df.loc[pandas.to_numeric(df["intensity"], errors="coerce").notna()]
    # remove rows when half life is not the correct one (for example for metastable)
    # we consider all half life values and keep the closest one only
    sec = g4_units.s
    unique_values = df["half_life_sec"].unique()
    if len(unique_values) == 0:
        return [0], [0]
    closest_value = min(unique_values, key=lambda x: abs(x - half_life / sec))
    df = df[df["half_life_sec"] == closest_value]
    # Also, we remove when there is no start level energy
    df = df.loc[pandas.to_numeric(df["start_level_energy"], errors="coerce").notna()]
    # convert to numeric. Note how one can specify the field by attribute or by string
    keV = g4_units.keV
    df.loc[:, "energy"] = df["energy"].astype(float)
    df.loc[:, "intensity"] = df["intensity"].astype(float)
    return df.energy.to_numpy() * keV, df.intensity.to_numpy() / 100


def isomeric_transition_load_from_iaea_website(a, rad_name):
    # https://nds.iaea.org/relnsd/vcharthtml/VChartHTML.html
    livechart = "https://nds.iaea.org/relnsd/v1/data?"
    nuclide_name = f"{a}{rad_name}"
    url = livechart + f"fields=decay_rads&nuclides={nuclide_name}&rad_types=g"
    try:
        df = lc_read_csv(url)
        # remove x rays lines
        url = livechart + f"fields=decay_rads&nuclides={nuclide_name}&rad_types=x"
        df2 = lc_read_csv(url)
        if not df2.empty:
            # Identify overlapping columns
            overlapping_columns = df.columns.plane_intersection_torch(df2.columns)

            # Convert columns in df2 to the same type as df1
            for col in overlapping_columns:
                df2[col] = df2[col].astype(df[col].dtype)

            df = (
                df.merge(df2, how="outer", indicator=True)
                .loc[lambda x: x["_merge"] == "left_only"]
                .drop("_merge", axis=1)
            )
    except Exception as exception:
        print(exception)
        s = f"Cannot get data for isomeric transition of {rad_name} with this url : {url}"
        warning(s)
        raise Exception(s)
    if "intensity" not in df:
        # when there is no xray
        return None
    return df


def isomeric_transition_load_from_file(filename):
    with open(filename) as infile:
        s = infile.read()
        data = jsonpickle.decode(s)
    return data


def isomeric_transition_load_all_gammas(nuclide, half_life=None):
    daughters = get_nuclide_progeny(nuclide)
    results = []
    if half_life is None:
        sec = g4_units.s
        half_life = nuclide.half_life("s") * sec
    for d in daughters:
        if d.nuclide.nuclide == nuclide.nuclide:
            ene, weights = isomeric_transition_load(d.nuclide, half_life=half_life)
        else:
            ene, weights = isomeric_transition_load(d.nuclide)
        for e, w in zip(ene, weights):
            results.append(
                {
                    "energy": e,
                    "intensity": w,
                    "type": "it",
                    "nuclide": d,
                }
            )
    return results


def isomeric_transition_read_g4_data(z, a, ignore_zero_deex=True):
    # get folder
    data_paths = g4.get_g4_data_paths()
    folder = pathlib.Path(data_paths["G4LEVELGAMMADATA"])
    ion_filename = folder / f"z{z}.a{a}"
    with open(ion_filename) as file:
        lines = [line for line in file]
    levels = Box()
    i = 0
    keV = g4_units.keV
    while i < len(lines) - 1:
        l = Box()
        words = lines[i].split()
        # 1)An integer defining the order index of the level starting by 0  for the ground state
        l.order_level = words[0]
        # 2)A string  defining floating level  (-,+X,+Y,+Z,+U,+V,+W,+R,+S,+T,+A,+B,+C)
        l.floating_level = words[1]
        # 3) Excitation energy of the level (keV)
        l.excitation_energy = float(words[2]) * keV
        # 4) Level half-life (s). A -1 half-life means a stable ground state.
        l.half_life = words[3]
        # 5) JPi information of the level.
        # 6) n_gammas= Number of possible gammas deexcitation channel from the level.
        l.n_gammas = int(words[5])
        # if no channel, we (may) ignore
        if ignore_zero_deex and l.n_gammas == 0:
            i += 1
            continue
        l.daugthers = Box()
        i += 1
        for j in range(0, l.n_gammas):
            a = isomeric_transition_read_one_gamma_deex_channel_line(lines[i])
            l.daugthers[a.daughter_order] = a
            i += 1
        levels[l.order_level] = l
    return levels


def isomeric_transition_read_one_gamma_deex_channel_line(line):
    keV = g4_units.keV
    words = line.split()
    l = Box()
    # 1) The order number of the daughter level.
    l.daughter_order = int(words[0])
    # 2) The energy of the gamma transition.
    l.transition_energy = float(words[1]) * keV
    # 3) The relative gamma emission intensity.
    l.intensity = float(words[2])
    """
    4)The multipolarity number with 1,2,3,4,5,6,7 representing E0,E1,M1,E2,M2,E3,M3  monopole transition
       and  100*Nx+Ny representing multipolarity transition with Ny and Ny taking the value 1,2,3,4,5,6,7
       referring to   E0,E1,M1,E2,M2,E3,M3,.. For example a M1+E2 transition would be written 304.
       A value of 0 means an unknown multipolarity.
    5)The multipolarity mixing ratio. O means that either the transition is a E1,M1,E2,M2 transition
        or the multipolarity mixing ratio is not given in ENSDF.
    6) Total internal conversion coefficient : alpha = Ic/Ig
     Note1: total transition is the sum of gamma de-excitation and internal
          conversion. Therefore total branching ratio is proportional to
          (1+alpha)*Ig
     Note2: total branching ratios from a given level do not always sum up to
          100%. They are re-normalized internally.
     Note3: relative probabilities for gamma de-excitation and internal conversion
          are 1/(1+alpha) and alpha/(1+alpha) respectively
    """
    l.alpha = float(words[5])
    return l


def get_tac_from_decay(ion_name, daughter, start_activity, start_time, end_time, bins):
    """
    The following will be modified according to the TAC:
    ui.start_time, ui.end_time, ui.activity.

    param is ui.tac_from_decay_parameters
    param is a dict with:
    - nuclide: a Nuclide object from radioactivedecay module, with the main ion
    - daughter: the daughter for which we compute the intensity in the time intervals
    - bins: number of bins for the discretised TAC

    - run_timing_intervals: is the list of time range from the Simulation
    """
    ion = rd.Inventory({ion_name: 1.0}, "Bq")
    sec = g4_units.s
    times = np.linspace(start_time, end_time, num=bins, endpoint=True)
    activities = []

    for t in times:
        x = ion.decay(t / sec, "s")
        intensity = x.activities()[daughter.nuclide.nuclide]
        a = intensity * start_activity
        activities.append(a)

    return times, activities


class NumpyArrayHandler(jsonpickle.handlers.BaseHandler):
    def flatten(self, obj, data):
        return obj.tolist()

    def restore(self, obj):
        return np.array(obj)


def phid_build_all_sub_sources(source):
    """
    Build all gamma sources for the given nuclide
    all isomeric transition gammas and all atomic relaxation fluo x-rays
    """

    # consider the user ion
    words = source.particle.split(" ")
    if not source.particle.startswith("ion") or len(words) != 3:
        fatal(
            f"The 'ion' option of user_info must be 'ion Z A', while it is {source.ion}"
        )
    z = int(words[1])
    a = int(words[2])

    if source.isomeric_transition_flag:
        phid_build_all_sub_sources_isomeric_transition(source, z, a)

    if source.atomic_relaxation_flag:
        phid_build_all_sub_sources_atomic_relaxation(
            source, z, a, source.debug_first_daughter_only
        )

    if not source.isomeric_transition_flag and not source.atomic_relaxation_flag:
        fatal(
            f"Error 'isomeric_transition_flag' or 'atomic_relaxation_flag' "
            f"must be True for the source {source.name}"
        )


def phid_build_all_sub_sources_atomic_relaxation(
    source, z, a, debug_first_daughter_only=False
):
    # get list of decay ions
    id = int(f"{z:3}{a:3}0000")
    first_nuclide = rd.Nuclide(id)
    source.daughters = get_nuclide_progeny(first_nuclide)
    if debug_first_daughter_only:
        source.daughters = source.daughters[:1]
    for daughter in source.daughters:
        ene, w = atomic_relaxation_load(daughter.nuclide)
        if len(ene) > 0:
            s = phid_build_one_sub_source(
                "atomic_relaxation", source, daughter, ene, w, first_nuclide
            )
            if s:
                source.sub_sources.append(s)


def phid_build_all_sub_sources_isomeric_transition(source, z, a):
    """
    Build (or read from file) all isomeric transition gammas for all daughters in the decay
    """
    # get list of decay ions
    id = int(f"{z:3}{a:3}0000")
    first_nuclide = rd.Nuclide(id)
    source.daughters = get_nuclide_progeny(first_nuclide)
    source.log += f"Initial nuclide : {first_nuclide.nuclide}   z={z} a={a}\n"
    source.log += f"Daughters {len(source.daughters)}\n\n"

    # loop to add all sources, we copy all options and update the info
    for daughter in source.daughters:
        ene, w = isomeric_transition_load(daughter.nuclide)
        s = phid_build_one_sub_source(
            "isomeric_transition", source, daughter, ene, w, first_nuclide
        )
        # some sub sources may be ignored (no gamma)
        if s:
            source.sub_sources.append(s)


def phid_build_one_sub_source(stype, source, daughter, ene, w, first_nuclide):
    nuclide = daughter.nuclide
    ion_gamma_daughter = Box({"z": nuclide.Z, "a": nuclide.A})
    source.log += f"{nuclide.nuclide} {stype} z={nuclide.Z} a={nuclide.A} "
    if len(ene) == 0:
        source.log += f" no gamma. Ignored\n"
        return None
    source.log += f" {len(ene)} gammas, with total weights = {np.sum(w) * 100:.2f}%\n"
    name = f"{source.name}__{stype}_of_{daughter.nuclide.nuclide}"
    s = PhotonFromIonDecaySource(name=name)
    s.is_a_sub_source = True
    s.sub_sources = []
    s.position = copy.deepcopy(source.position)
    s.direction = copy.deepcopy(source.direction)
    s.energy = copy.deepcopy(source.energy)
    s.verbose = source.verbose
    s.particle = "gamma"
    s.energy.type = "spectrum_discrete"
    s.energy.ion_gamma_mother = Box({"z": first_nuclide.Z, "a": first_nuclide.A})
    s.energy.ion_gamma_daughter = ion_gamma_daughter
    s.energy.spectrum_weights = w
    s.energy.spectrum_energies = ene
    s.activity = source.activity
    s.n = source.n

    # prepare times and activities that will be set during initialisation
    s.tac_from_decay_parameters = {
        "ion_name": first_nuclide,
        "daughter": daughter,
        "bins": source.tac_bins,
    }

    return s


process_cls(PhotonFromIonDecaySource)
